# Copyright cocotb contributors
# Licensed under the Revised BSD License, see LICENSE for details.
# SPDX-License-Identifier: BSD-3-Clause
#!/usr/bin/env python
"""
Simple script to combine JUnit test results into a single XML file.
"""

import argparse
import os
import re
import sys
from pathlib import Path
from typing import Iterable, Pattern
from xml.etree import ElementTree as ET


def _find_all(name: Pattern, path: Path) -> Iterable[Path]:
    for obj in path.iterdir():
        if obj.is_file() and re.match(name, obj.name):
            yield obj
        elif obj.is_dir():
            yield from _find_all(name, obj)


def _get_parser() -> argparse.ArgumentParser:
    """Return the cmdline parser"""
    parser = argparse.ArgumentParser(
        description=__doc__, formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        "directories",
        nargs="*",
        type=lambda args: [Path(arg) for arg in args],
        default=[Path()],
        help="Directories to search for input files.",
    )
    parser.add_argument(
        "-i",
        "--input-filename",
        default=r"results.*\.xml",
        help="A regular expression to match input filenames.",
    )
    parser.add_argument(
        "-o",
        "--output-file",
        default="combined_results.xml",
        help="Path of output XML file.",
    )
    parser.add_argument(
        "--output-testsuites-name",
        default="results",
        help="Name of 'testsuites' element in output XML file.",
    )
    parser.add_argument(
        "--verbose",
        action="store_true",
        help="Enables verbose output.",
    )
    parser.add_argument(
        "--repo-root",
        type=Path,
        help="Specify root of cocotb repo the regression is run from (CI only).",
    )
    return parser


def main() -> int:
    parser = _get_parser()
    args = parser.parse_args()
    rc = 0

    result = ET.Element("testsuites", name=args.output_testsuites_name)

    input_pattern = re.compile(args.input_filename)

    for directory in args.directories:
        if args.verbose:
            print(f"Searching in {directory} for results.xml files.")
        for fname in _find_all(input_pattern, directory):
            if args.verbose:
                print(f"Reading file {fname}.")
            tree = ET.parse(fname)
            for ts in tree.iter("testsuite"):
                if args.verbose:
                    print(
                        "Testsuite name: {!r}, package: {!r}".format(
                            ts.get("name"), ts.get("package")
                        )
                    )
                for existing in result:
                    if (existing.get("name") == ts.get("name")) and (
                        existing.get("package") == ts.get("package")
                    ):
                        if args.verbose:
                            print(
                                "Testsuite already exists in combined results. Extending it."
                            )
                        existing.extend(list(ts))
                        break
                else:
                    if args.verbose:
                        print(
                            "Testsuite does not already exist in combined results. Adding it."
                        )
                    result.append(ts)

    testsuite_count = 0
    testcase_count = 0
    for testsuite in result.iter("testsuite"):
        testsuite_count += 1
        for testcase in testsuite.iter("testcase"):
            testcase_count += 1
            for _ in testcase.iter("failure"):
                rc = 1
                print(
                    "Failure in testsuite: '{}' classname: '{}' testcase: '{}' with parameters '{}'".format(
                        testsuite.get("name"),
                        testcase.get("classname"),
                        testcase.get("name"),
                        testsuite.get("package"),
                    )
                )
                if (
                    os.getenv("GITHUB_ACTIONS") is not None
                    and args.repo_root is not None
                ):
                    # Get test file relative to root of repo
                    file = testcase.get("file")
                    # if this file was output by cocotb, it has this attribute
                    assert file is not None
                    relative_file = Path(file).relative_to(args.repo_root)
                    print(
                        "::error file={},line={}::Test {}:{} failed".format(
                            relative_file,
                            testcase.get("lineno"),
                            testcase.get("classname"),
                            testcase.get("name"),
                        )
                    )

    print(f"Ran a total of {testsuite_count} TestSuites and {testcase_count} TestCases")

    if args.verbose:
        print(f"Writing combined results to {args.output_file}")
    ET.ElementTree(result).write(args.output_file, encoding="UTF-8")
    return rc


if __name__ == "__main__":
    rc = main()
    sys.exit(rc)



# Copyright cocotb contributors
# Licensed under the Revised BSD License, see LICENSE for details.
# SPDX-License-Identifier: BSD-3-Clause
import warnings


def __getattr__(name: str) -> object:
    if name == "TestSuccess":
        warnings.warn(
            "`raise TestSuccess` is deprecated. Use `cocotb.pass_test()` instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        from cocotb._test_functions import TestSuccess  # noqa: PLC0415

        return TestSuccess

    elif name == "SimFailure":
        warnings.warn(
            "SimFailure was moved to `cocotb.regression`.",
            DeprecationWarning,
            stacklevel=2,
        )
        from cocotb.regression import SimFailure  # noqa: PLC0415

        return SimFailure

    elif name == "SimTimeoutError":
        warnings.warn(
            "SimTimeoutError was moved from `cocotb.result` to `cocotb.triggers`.",
            DeprecationWarning,
            stacklevel=2,
        )
        from cocotb.triggers import SimTimeoutError  # noqa: PLC0415

        return SimTimeoutError

    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



# Copyright cocotb contributors
# Licensed under the Revised BSD License, see LICENSE for details.
# SPDX-License-Identifier: BSD-3-Clause

# generated with mypy's stubgen script

from logging import Logger
from typing import Any, Callable

from cocotb.handle import GPIDiscovery

DRIVERS: int
ENUM: int
GENARRAY: int
INTEGER: int
LOADS: int
LOGIC: int
LOGIC_ARRAY: int
MEMORY: int
MODULE: int
NETARRAY: int
OBJECTS: int
PACKAGE: int
REAL: int
STRING: int
STRUCTURE: int
PACKED_STRUCTURE: int
UNKNOWN: int
RISING: int
FALLING: int
VALUE_CHANGE: int
RANGE_UP: int
RANGE_DOWN: int
RANGE_NO_DIR: int

class gpi_cb_hdl:
    def deregister(self) -> None: ...
    def __eq__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...
    def __hash__(self) -> int: ...

class gpi_iterator_hdl:
    def __eq__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...
    def __hash__(self) -> int: ...
    def __iter__(self) -> gpi_iterator_hdl: ...
    def __next__(self) -> gpi_sim_hdl: ...

class gpi_sim_hdl:
    def get_const(self) -> bool: ...
    def get_definition_file(self) -> str: ...
    def get_definition_name(self) -> str: ...
    def get_handle_by_index(self, index: int) -> gpi_sim_hdl | None: ...
    def get_handle_by_name(
        self, name: str, discovery_method: GPIDiscovery | None = GPIDiscovery.AUTO
    ) -> gpi_sim_hdl | None: ...
    def get_indexable(self) -> bool: ...
    def get_name_string(self) -> str: ...
    def get_num_elems(self) -> int: ...
    def get_range(self) -> tuple[int, int, int]: ...
    def get_signal_val_binstr(self) -> str: ...
    def get_signal_val_long(self) -> int: ...
    def get_signal_val_real(self) -> float: ...
    def get_signal_val_str(self) -> bytes: ...
    def get_type(self) -> int: ...
    def get_type_string(self) -> str: ...
    def iterate(self, mode: int) -> gpi_iterator_hdl: ...
    def set_signal_val_binstr(self, action: int, value: str) -> None: ...
    def set_signal_val_int(self, action: int, value: int) -> None: ...
    def set_signal_val_real(self, action: int, value: float) -> None: ...
    def set_signal_val_str(self, action: int, value: bytes) -> None: ...
    def __eq__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...
    def __hash__(self) -> int: ...

def get_precision() -> int: ...
def get_root_handle(name: str | None) -> gpi_sim_hdl | None: ...
def get_sim_time() -> tuple[int, int]: ...
def get_simulator_product() -> str: ...
def get_simulator_version() -> str: ...
def is_running() -> bool: ...
def set_gpi_log_level(level: int) -> None: ...
def package_iterate() -> gpi_iterator_hdl: ...
def register_nextstep_callback(func: Callable[..., Any], *args: Any) -> gpi_cb_hdl: ...
def register_readonly_callback(func: Callable[..., Any], *args: Any) -> gpi_cb_hdl: ...
def register_rwsynch_callback(func: Callable[..., Any], *args: Any) -> gpi_cb_hdl: ...
def register_timed_callback(
    time: int, func: Callable[..., Any], *args: Any
) -> gpi_cb_hdl: ...
def register_value_change_callback(
    signal: gpi_sim_hdl, func: Callable[..., Any], edge: int, *args: Any
) -> gpi_cb_hdl: ...
def stop_simulator() -> None: ...

class cpp_clock:
    def __init__(self, signal: gpi_sim_hdl) -> None: ...
    def start(
        self, period_steps: int, high_steps: int, start_high: bool, set_action: int
    ) -> None: ...
    def stop(self) -> None: ...

def clock_create(hdl: gpi_sim_hdl) -> cpp_clock: ...
def initialize_logger(
    log_func: Callable[[Logger, int, str, int, str, str], None],
    get_logger: Callable[[str], Logger],
) -> None: ...
def set_sim_event_callback(sim_event_callback: Callable[[str], None]) -> None: ...



# Copyright cocotb contributors
# Licensed under the Revised BSD License, see LICENSE for details.
# SPDX-License-Identifier: BSD-3-Clause
from abc import abstractmethod
from typing import Generic, Iterable, Iterator, Optional, TypeVar, Union, overload

from cocotb.types._range import Range

T_co = TypeVar("T_co", covariant=True)
T = TypeVar("T")


class AbstractArray(Generic[T_co]):
    r"""Abstract base class for non-mutating Array-like collections.

    Arrays are similar to :class:`~collections.abc.Sequence`\ s,
    but their size cannot change after creation and they support arbitrary indexing schemes.

    Abstract methods
    ^^^^^^^^^^^^^^^^

    * :attr:`range`
    * :meth:`!__getitem__`

    Mixin methods
    ^^^^^^^^^^^^^

    * :attr:`left`, :attr:`right`, and :attr:`direction`
    * :meth:`!__len__`
    * :meth:`!__iter__` and :meth:`!__reversed__`
    * :meth:`!__contains__`
    * :meth:`index` and :meth:`count`
    """

    @property
    def left(self) -> int:
        """Leftmost index of the array."""
        return self.range.left

    @property
    def direction(self) -> str:
        """``"to"`` if indexes are ascending, ``"downto"`` otherwise."""
        return self.range.direction

    @property
    def right(self) -> int:
        """Rightmost index of the array."""
        return self.range.right

    @property
    @abstractmethod
    def range(self) -> Range:
        """:class:`Range` of the indexes of the array."""

    @range.setter
    @abstractmethod
    def range(self, new_range: Range) -> None:
        """Set a new indexing scheme on the array.

        Must be the same size.
        """

    def __len__(self) -> int:
        return len(self.range)

    def __iter__(self) -> Iterator[T_co]:
        for i in self.range:
            yield self[i]

    def __reversed__(self) -> Iterator[T_co]:
        for i in reversed(self.range):
            yield self[i]

    def __contains__(self, item: object) -> bool:
        return any(v == item for v in self)

    @overload
    def __getitem__(self, item: int) -> T_co: ...

    @overload
    def __getitem__(self, item: slice) -> "AbstractArray[T_co]": ...

    @abstractmethod
    def __getitem__(
        self, item: Union[int, slice]
    ) -> Union[T_co, "AbstractArray[T_co]"]: ...

    def index(
        self,
        value: object,
        start: Optional[int] = None,
        stop: Optional[int] = None,
    ) -> int:
        """Find first occurrence of value.

        Args:
            value: Value to search for.
            start: Index to start search at.
            stop: Index to stop search at.

        Returns:
            Index of first occurrence of *value*.

        Raises:
            ValueError: If the value is not present.
        """
        if start is None:
            start = self.left
        if stop is None:
            stop = self.right
        for i in Range(start, self.direction, stop):
            if self[i] == value:
                return i
        raise IndexError(f"{value!r} not in array")

    def count(self, value: object) -> int:
        """Return number of occurrences of value.

        Args:
            value: Value to search for.

        Returns:
            Number of occurrences of *value*.
        """
        count: int = 0
        for v in self:
            if v == value:
                count += 1
        return count


class AbstractMutableArray(AbstractArray[T]):
    """Abstract base class for mutating Array-like collections.

    See :class:`.AbstractArray` for more details.

    Additional abstract methods:

    * :meth:`!__setitem__`
    """

    @overload
    def __setitem__(self, item: int, value: T) -> None: ...

    @overload
    def __setitem__(self, item: slice, value: Iterable[T]) -> None: ...

    @abstractmethod
    def __setitem__(
        self, item: Union[int, slice], value: Union[T, Iterable[T]]
    ) -> None: ...



# Copyright cocotb contributors
# Licensed under the Revised BSD License, see LICENSE for details.
# SPDX-License-Identifier: BSD-3-Clause


class InternalError(BaseException):
    """An error internal to scheduler. If you see this, report a bug!"""



# Copyright cocotb contributors
# Licensed under the Revised BSD License, see LICENSE for details.
# SPDX-License-Identifier: BSD-3-Clause


# Debug mode controlled by environment variables
import cProfile
import os
import pstats

from cocotb._py_compat import AbstractContextManager, nullcontext

profiling_context: AbstractContextManager[None, None]


if "COCOTB_ENABLE_PROFILING" in os.environ:
    _profile: cProfile.Profile

    def initialize() -> None:
        global _profile
        _profile = cProfile.Profile()

    def finalize() -> None:
        ps = pstats.Stats(_profile).sort_stats("cumulative")
        ps.dump_stats("cocotb.pstat")

    class _profiling_context(AbstractContextManager[None, None]):
        """Context manager that profiles its contents"""

        def __enter__(self) -> None:
            _profile.enable()

        def __exit__(self, *excinfo: object) -> None:
            _profile.disable()

    profiling_context = _profiling_context()

else:

    def initialize() -> None:
        pass

    def finalize() -> None:
        pass

    profiling_context = nullcontext()



# Copyright cocotb contributors
# Licensed under the Revised BSD License, see LICENSE for details.
# SPDX-License-Identifier: BSD-3-Clause
"""Collection of functions to control a running test and related exceptions."""

from typing import NoReturn, Type, Union

Failed: Type[BaseException]
try:
    import pytest
except ModuleNotFoundError:
    Failed = AssertionError
else:
    try:
        with pytest.raises(Exception):
            pass
    except BaseException as _raises_e:
        Failed = type(_raises_e)
    else:
        assert False, "pytest.raises doesn't raise an exception when it fails"


class TestSuccess(BaseException):
    """Implementation of :func:`pass_test`.

    Users are *not* intended to catch this exception type.
    """

    def __init__(self, msg: Union[str, None]) -> None:
        super().__init__(msg)
        self.msg = msg


def pass_test(msg: Union[str, None] = None) -> NoReturn:
    """Force a test to pass.

    The test will end and enter termination phase when this is called.

    Args:
        msg: The message to display when the test passes.
    """
    raise TestSuccess(msg)



# Copyright cocotb contributors
# Licensed under the Revised BSD License, see LICENSE for details.
# SPDX-License-Identifier: BSD-3-Clause
from cocotb._py_compat import Literal, TypeAlias

RoundMode: TypeAlias = Literal["error", "round", "ceil", "floor"]
TimeUnit: TypeAlias = Literal["step", "fs", "ps", "ns", "us", "ms", "sec"]



# Copyright cocotb contributors
# Licensed under the Revised BSD License, see LICENSE for details.
# SPDX-License-Identifier: BSD-3-Clause
import os
from functools import lru_cache
from random import getrandbits
from typing import Callable, Dict, Union

from cocotb._py_compat import Final, Literal, TypeAlias

ResolverLiteral: TypeAlias = Literal["weak", "zeros", "ones", "random"]


_ord_0 = ord("0")


class _random_resolve_table(Dict[int, int]):
    def __init__(self) -> None:
        self[ord("0")] = ord("0")
        self[ord("1")] = ord("1")
        self[ord("L")] = ord("0")
        self[ord("H")] = ord("1")

    def __missing__(self, _: str) -> int:
        return getrandbits(1) + _ord_0


_resolve_tables: Dict[str, Dict[int, int]] = {
    "error": {},
    "weak": str.maketrans("LHW", "01X"),
    "zeros": str.maketrans("LHUXZW-", "0100000"),
    "ones": str.maketrans("LHUXZW-", "0111111"),
    "random": _random_resolve_table(),
}

_VALID_RESOLVERS = ("error", "weak", "zeros", "ones", "random")
_VALID_RESOLVERS_ERR_MSG = (
    "Valid values are 'error', 'weak', 'zeros', 'ones', or 'random'"
)


@lru_cache(maxsize=None)
def get_str_resolver(resolver: ResolverLiteral) -> Callable[[str], str]:
    if resolver not in _VALID_RESOLVERS:
        raise ValueError(f"Invalid resolver: {resolver!r}. {_VALID_RESOLVERS_ERR_MSG}")

    resolve_table = _resolve_tables[resolver]

    def resolve_func(value: str) -> str:
        return value.translate(resolve_table)

    return resolve_func


def _init() -> Union[Callable[[str], str], None]:
    _envvar = os.getenv("COCOTB_RESOLVE_X", None)

    # no resolver
    if _envvar is None:
        return None

    # backwards compatibility
    resolver = _envvar.strip().lower()
    if resolver == "value_error":
        resolver = "error"

    # get resolver
    try:
        return get_str_resolver(resolver)
    except ValueError:
        raise ValueError(
            f"Invalid COCOTB_RESOLVE_X value: {_envvar!r}. {_VALID_RESOLVERS_ERR_MSG}"
        ) from None


RESOLVE_X: Final = _init()



# Copyright cocotb contributors
# Licensed under the Revised BSD License, see LICENSE for details.
# SPDX-License-Identifier: BSD-3-Clause
import copy
import warnings
from math import ceil
from typing import (
    Any,
    Dict,
    Iterable,
    Iterator,
    List,
    Union,
    cast,
    overload,
)

from cocotb._deprecation import deprecated
from cocotb._py_compat import Literal, TypeAlias
from cocotb.types._abstract_array import AbstractMutableArray
from cocotb.types._indexing import IndexingChangedWarning
from cocotb.types._logic import Logic, LogicConstructibleT
from cocotb.types._range import Range
from cocotb.types._resolve import RESOLVE_X, ResolverLiteral, get_str_resolver

_resolve_lh_table = str.maketrans({"L": "0", "H": "1"})
_str_literals = frozenset("uUxX01zZwWlLhH-")


ByteOrder: TypeAlias = Literal["big", "little"]


class LogicArray(AbstractMutableArray[Logic]):
    r"""Fixed-sized, arbitrarily-indexed, :class:`.Array` of :class:`.Logic`\ s.

    An :class:`!Array`, where all elements are enforced to be :class:`!Logic`.
    Additionally, this type supports bit-wise logical operators, conversions to integers and bytes,
    and ``X`` testing and mapping.

    :class:`!LogicArray`\ s can be constructed from an iterable of :class:`!Logic`\ s,
    or values constructible into :class:`!Logic`, like :class:`bool`, :class:`str`, or :class:`int`.
    Alternatively, they can be constructed from :class:`!str` or :class:`!int` literals.

    Like :class:`!Array`, if *range* is not given, the range ``Range(len(value)-1, "downto", 0)`` is used;
    and if an :class:`int` is passed for *range*, the range ``Range(range-1, "downto", 0)`` is used.

    .. code-block:: pycon3

        >>> LogicArray(0b0111, 4)
        LogicArray('0111', Range(3, 'downto', 0))

        >>> LogicArray("01XZ", Range(0, "to", 3))
        LogicArray('01XZ', Range(0, 'to', 3))

        >>> LogicArray([0, True, "X", Logic("-")])
        LogicArray('01X-', Range(3, 'downto', 0))

    .. note::
        If constructing from an unsigned :class:`!int` literal, *range* `must` be given.

    :class:`!LogicArray`\ s can be constructed from :class:`int`\ s using :meth:`from_unsigned` or :meth:`from_signed`.

    .. code-block:: pycon3

        >>> LogicArray.from_unsigned(0xA, 4)
        LogicArray('1010', Range(3, 'downto', 0))

        >>> LogicArray.from_signed(-4, Range(0, "to", 3))  # will sign-extend
        LogicArray('1100', Range(0, 'to', 3))

    :class:`!LogicArray`\ s can be constructed from :class:`bytes` or :class:`bytearray` using :meth:`from_bytes`.
    Use the *byteorder* argument to control endianness.

    .. code-block:: pycon3

        >>> LogicArray.from_bytes(b"1n", byteorder="big")
        LogicArray('0011000101101110', Range(15, 'downto', 0))

        >>> LogicArray.from_bytes(b"1n", byteorder="little")
        LogicArray('0110111000110001', Range(15, 'downto', 0))

    :class:`!LogicArray`\ s support the same :class:`list`-like operations as :class:`!Array`;
    however, it enforces the condition that all elements must be a :class:`!Logic`.

    .. code-block:: pycon3

        >>> array = LogicArray("1010")
        >>> array[0]  # is indexable
        Logic('0')

        >>> array[1:]  # is slice-able
        LogicArray('10', Range(1, 'downto', 0))

        >>> Logic("0") in array  # is a collection
        True

        >>> list(array)  # is an iterable
        [Logic('1'), Logic('0'), Logic('1'), Logic('0')]

    When setting an element or slice, the *value* is first constructed into a :class:`!Logic`.

    .. code-block:: pycon3

        >>> array = LogicArray("1010")
        >>> array[3] = "Z"
        >>> array[3]
        Logic('Z')

        >>> array[2:] = ["X", True, 0]
        >>> array
        LogicArray('ZX10', Range(3, 'downto', 0))

        >>> array[:] = 0b0101
        >>> array
        LogicArray('0101', Range(3, 'downto', 0))

    :class:`!LogicArray`\ s can be converted into their :class:`str` or :class:`int` literal values using casts.
    They can also be used in conditionals.

    .. code-block:: pycon3

        >>> value = LogicArray("1010")
        >>> str(value)
        '1010'

        >>> int(value)
        10
        >>> if value:
        ...     print("Not 0!")
        Not 0!

    .. warning::
        The :class:`int` cast, :class:`bool` cast, and use in conditionals assumes the
        value is entirely ``0``, ``1``, ``L``, or ``H``, and will raise an exception otherwise.

    The :meth:`to_unsigned`, :meth:`to_signed`, and :meth:`to_bytes` methods can be used to convert
    the value into an unsigned or signed integer, or bytes, respectively.

    .. code-block:: pycon3

        >>> value = LogicArray("1010")
        >>> value.to_unsigned()
        10

        >>> value.to_signed()
        -6

        >>> value.to_bytes(byteorder="big")
        b'\n'

    .. warning::
        These operations assume the value is entirely ``0``, ``1``, ``L``, or ``H``, and will raise an exception otherwise.

    You can also convert :class:`!LogicArray`\ s to hexadecimal or binary strings using
    the built-ins :func:`hex` and :func:`bin`, respectively.

    .. code-block:: pycon3

        >>> value = LogicArray("01111010")
        >>> hex(value)
        '0x7a'

        >>> bin(value)
        '0b1111010'

    .. warning::
        Using :func:`hex` or :func:`bin` first turns the :class:`!LogicArray` into an :class:`int`.
        This means the exact length of the :class:`!LogicArray` is lost.
        It also means that these expressions will raise an exception if the value is not entirely ``0``, ``1``, ``L``, or ``H``.

    :class:`!LogicArray`\ s also support element-wise logical operations: ``&``, ``|``,
    ``^``, and ``~``.

    .. code-block:: pycon3

        >>> def big_mux(a: LogicArray, b: LogicArray, sel: Logic) -> LogicArray:
        ...     s = LogicArray([sel] * len(a))
        ...     return (a & ~s) | (b & s)

        >>> a = LogicArray("0110")
        >>> b = LogicArray("1110")
        >>> sel = Logic("1")  # choose second option
        >>> big_mux(a, b, sel)
        LogicArray('1110', Range(3, 'downto', 0))

    Args:
        value: Initial value for the :class:`!LogicArray`.
        range: The indexing scheme of the :class:`!LogicArray`.

    Raises:
        TypeError: When invalid argument types are used.
        ValueError: When *value* will not fit in a :class:`!LogicArray` of the given *range*.
    """

    # These three attribute contain the current value of the array in one or more of
    # three different implementations. This is done for performance reasons, as certain
    # implementations are faster for particular operations.
    # Each implementation can be present, or None if the implementation has not been
    # computed or has been invalidated by a mutating operation.
    _value_as_array: Union[List[Logic], None]
    _value_as_int: Union[int, None]
    _value_as_str: Union[str, None]
    _range: Range
    _warn_indexing: bool

    __slots__ = (
        "_value_as_array",
        "_value_as_int",
        "_value_as_str",
        "_range",
        "_warn_indexing",
    )

    def __init__(
        self,
        value: Union[int, str, Iterable[LogicConstructibleT]],
        range: Union[Range, int, None] = None,
    ) -> None:
        self._value_as_array = None
        self._value_as_int = None
        self._value_as_str = None
        self._warn_indexing = False

        if isinstance(range, int):
            range = Range(range - 1, "downto", 0)
        elif range is not None and not isinstance(range, Range):
            raise TypeError(
                f"Expected Range or int for parameter 'range', not {type(range).__qualname__}"
            )

        if isinstance(value, str):
            if not (set(value) <= _str_literals):
                raise ValueError("Invalid str literal")
            self._value_as_str = value.upper()
            if range is not None:
                if len(value) != len(range):
                    raise ValueError(f"Length of {value!r} does not match {range!r}")
                self._range = range
            else:
                self._range = Range(len(self._value_as_str) - 1, "downto", 0)
        elif isinstance(value, int):
            value = int(value)  # force bool to int
            if value < 0:
                raise ValueError("Invalid int literal")
            if range is None:
                raise TypeError("Missing required arguments: 'range'")
            bitlen = max(1, int.bit_length(value))
            if bitlen > len(range):
                raise ValueError(
                    f"{value!r} is too large to fit in LogicArray with {range!r}"
                )
            self._value_as_int = value
            self._range = range
        elif isinstance(value, LogicArray):
            array = value._value_as_array
            self._value_as_array = list(array) if array is not None else None
            self._value_as_int = value._value_as_int
            self._value_as_str = value._value_as_str
            if range is None:
                self._range = value._range
            else:
                if len(range) != len(value):
                    raise ValueError(f"Length of {value!r} does not match {range!r}")
                self._range = range
        else:
            self._value_as_array = [Logic(v) for v in value]
            if range is not None:
                if len(self._value_as_array) != len(range):
                    raise ValueError(
                        f"Value of length {len(self._value_as_array)} does not match range: {range!r}"
                    )
                self._range = range
            else:
                self._range = Range(len(self._value_as_array) - 1, "downto", 0)

    def _get_array(self) -> List[Logic]:
        if self._value_as_array is None:
            # May convert int to str before to converting to array.
            self._value_as_array = [Logic(v) for v in self._get_str()]
        return self._value_as_array

    def _get_str(self) -> str:
        if self._value_as_str is None:
            if self._value_as_int is not None:
                self._value_as_str = format(self._value_as_int, f"0{len(self)}b")
            else:
                self._value_as_str = "".join(
                    str(v) for v in cast("List[Logic]", self._value_as_array)
                )
        return self._value_as_str

    def _get_int(self) -> int:
        if self._value_as_int is None:
            # May convert list to str before converting to int.
            value_as_str = self._get_str()

            # always resolve L and H to 0 and 1
            value_as_str = value_as_str.translate(_resolve_lh_table)

            try:
                self._value_as_int = int(value_as_str, 2)
            except ValueError:
                if RESOLVE_X is None:
                    raise ValueError(
                        f"Can't convert {type(self).__qualname__} to int: it contains non-0/1 values"
                    ) from None
                else:
                    value_as_str = RESOLVE_X(value_as_str)
                    return int(value_as_str, 2)

        return self._value_as_int

    @classmethod
    def from_unsigned(
        cls,
        value: int,
        range: Union[Range, int],
    ) -> "LogicArray":
        """Construct a :class:`!LogicArray` from an :class:`int` with unsigned representation.

        The :class:`int` is treated as an arbitrary-length bit vector with unsigned representation where the left-most bit is the most significant bit.
        This bit vector is then constructed into a :class:`!LogicArray`.

        Args:
            value: The integer to convert.
            range: Indexing scheme for the :class:`!LogicArray`.

        Returns:
            A :class:`!LogicArray` equivalent to the *value*.

        Raises:
            TypeError: When invalid argument types are used.
            ValueError: When a :class:`!LogicArray` of the given *range* can't hold the *value*, or *value* is negative.
        """
        if value < 0:
            raise ValueError("Expected unsigned integer, got negative value")
        return LogicArray(value, range)

    @classmethod
    def from_signed(
        cls,
        value: int,
        range: Union[Range, int],
    ) -> "LogicArray":
        """Construct a :class:`!LogicArray` from an :class:`int` with two's complement representation.

        The :class:`int` is treated as an arbitrary-length bit vector with two's complement representation where the left-most bit is the most significant bit.
        This bit vector is then constructed into a :class:`!LogicArray`.

        Args:
            value: The integer to convert.
            range: Indexing scheme for the :class:`!LogicArray`.

        Returns:
            A :class:`!LogicArray` equivalent to the *value*.

        Raises:
            TypeError: When invalid argument types are used.
            ValueError: When a :class:`!LogicArray` of the given *range* can't hold the *value*.
        """
        if isinstance(range, int):
            range = Range(range - 1, "downto", 0)
        elif not isinstance(range, Range):
            raise TypeError(
                f"Expected Range or int for parameter 'range', not {type(range).__qualname__}"
            )

        # Prevent null range from blowing up the below code.
        if len(range) == 0:
            raise ValueError(
                f"Signed integer {value!r} will not fit in a LogicArray with bounds: {range!r}"
            )

        limit = 1 << (len(range) - 1)
        if value < -limit or limit <= value:
            raise ValueError(
                f"Signed integer {value!r} will not fit in a LogicArray with bounds: {range!r}"
            )
        value %= 2 * limit

        return LogicArray(value, range)

    @classmethod
    def from_bytes(
        cls,
        value: Union[bytes, bytearray],
        range: Union[Range, int, None] = None,
        *,
        byteorder: ByteOrder,
    ) -> "LogicArray":
        """Construct a :class:`!LogicArray` from :class:`bytes`.

        The :class:`bytes` is first converted to an unsigned integer using *byteorder*-endian representation,
        then is converted to a :class:`!LogicArray` as in :meth:`from_unsigned`.

        Args:
            value: The bytes to convert.
            range: Indexing scheme for the :class:`!LogicArray`.
            byteorder: The endianness used to construct the intermediate integer, either ``"big"`` or ``"little"``.

        Returns:
            A :class:`!LogicArray` equivalent to the *value*.

        Raises:
            ValueError: When a :class:`!LogicArray` of the given *range* can't hold the *value*.
        """
        if range is None:
            range = Range(len(value) * 8 - 1, "downto", 0)
        else:
            if isinstance(range, int):
                range = Range(range - 1, "downto", 0)
            if len(value) * 8 != len(range):
                raise ValueError(
                    f"Value of length {len(value)} will not fit in a LogicArray with bounds: {range!r}"
                )
        value_as_int = int.from_bytes(value, byteorder=byteorder, signed=False)
        return LogicArray(value_as_int, range)

    @classmethod
    def _from_handle(cls, value: str, warn_indexing: bool) -> "LogicArray":
        # Used by cocotb.handle classes to make LogicArray from values gotten from the
        # simulator which we expect to be well-formed.
        # Values are required to be uppercase.
        self = cls.__new__(cls)
        self._value_as_array = None
        self._value_as_int = None
        self._value_as_str = value
        self._range = Range(len(value) - 1, "downto", 0)
        self._warn_indexing = warn_indexing
        return self

    @property
    def range(self) -> Range:
        """:class:`Range` of the indexes of the array."""
        return self._range

    @range.setter
    def range(self, new_range: Range) -> None:
        """Set a new indexing scheme on the array. Must be the same size."""
        if not isinstance(new_range, Range):
            raise TypeError("range argument must be of type 'Range'")
        if len(new_range) != len(self):
            raise ValueError(
                f"{new_range!r} not the same length as old range: {self._range!r}"
            )
        self._range = new_range

    def __iter__(self) -> Iterator[Logic]:
        return iter(self._get_array())

    def __reversed__(self) -> Iterator[Logic]:
        return reversed(self._get_array())

    def __contains__(self, item: object) -> bool:
        return item in self._get_array()

    def __eq__(
        self,
        other: object,
    ) -> bool:
        if isinstance(other, int):
            if len(self) == 0:
                # Null arrays don't have a value and thus always compare False.
                return False
            try:
                return self.to_unsigned() == other
            except ValueError:
                return False
        elif isinstance(other, str):
            return str(self) == other.upper()
        elif isinstance(other, LogicArray):
            if len(self) != len(other):
                return False
            # Complex, but efficient chain of checking logic.
            # Avoid conversions if it can help it at first.
            # Prefers checking against str vs any type since that is going to be the
            #   most common type and also the "middle" type for conversions.
            # Always converts away from ints to prevent issues with non-0/1 data.
            if self._value_as_str is not None and other._value_as_str is not None:
                # (STR, STR)
                return self._value_as_str == other._value_as_str
            elif self._value_as_array is not None and other._value_as_array is not None:
                # (ARRAY, ARRAY)
                return self._value_as_array == other._value_as_array
            elif self._value_as_int is not None and other._value_as_int is not None:
                # (INT, INT)
                return self._value_as_int == other._value_as_int
            elif self._value_as_str is not None:
                # (STR, INT)
                # (STR, ARRAY)
                return self._value_as_str == other._get_str()
            elif other._value_as_str is not None:
                # (INT, STR)
                # (ARRAY, STR)
                return self._get_str() == other._value_as_str
            elif self._value_as_array is not None:
                # (ARRAY, INT)
                return self._value_as_array == other._get_array()
            else:
                # (INT, ARRAY)
                return self._get_array() == other._value_as_array
        elif isinstance(other, (list, tuple)):
            try:
                other = LogicArray(other)
            except ValueError:
                return False
            return self == other
        else:
            return NotImplemented

    @property
    @deprecated(
        "`logic_array.binstr` getter is deprecated. Use `str(logic_array)` instead."
    )
    def binstr(self) -> str:
        """The :class:`!LogicArray`'s value in :class:`str` literal representation.

        :getter:
            Return the :class:`!LogicArray`'s value in :class:`str` literal representation.

            .. deprecated:: 2.0
                Use ``str(logic_array)`` instead.

        :setter:
            Set the :class:`!LogicArray`'s value using a :class:`str` literal representation.

            .. deprecated:: 2.0
                Use ``logic_array[:] = value`` instead.
        """
        return str(self)

    @binstr.setter
    @deprecated(
        "`logic_array.binstr = value` setter is deprecated. Use `logic_array[:] = value` instead."
    )
    def binstr(self, value: str) -> None:
        self[:] = value

    @property
    def is_resolvable(self) -> bool:
        """``True`` if all elements are ``0``, ``1``, ``L``, ``H``."""
        return all(bit.is_resolvable for bit in self)

    @property
    @deprecated(
        "`logic_array.integer` getter is deprecated. Use `logic_array.to_unsigned()` instead."
    )
    def integer(self) -> int:
        """The :class:`!LogicArray`'s value as an unsigned :class:`int`.

        The :class:`!LogicArray` is treated as an arbitrary-length vector of bits
        with the left-most bit being the most significant bit in the integer value.
        The bit vector is then interpreted as an integer using unsigned representation.

        :getter:
            Return the :class:`!LogicArray`'s value as an unsigned integer.

            .. deprecated:: 2.0
                Use :meth:`logic_array.to_unsigned() <cocotb.types.LogicArray.to_unsigned>` instead.

        :setter:
            Set the :class:`!LogicArray`'s value using an unsigned integer.

            .. deprecated:: 2.0
                Use ``logic_array[:] = value`` instead.
        """
        return self.to_unsigned()

    @integer.setter
    @deprecated(
        "`logic_array.integer = value` setter is deprecated. Use `logic_array[:] = value` instead."
    )
    def integer(self, value: int) -> None:
        self[:] = value

    @property
    @deprecated(
        "`logic_array.signed_integer` getter is deprecated. Use `logic_array.to_signed()` instead."
    )
    def signed_integer(self) -> int:
        """The :class:`!LogicArray`'s value as a signed :class:`int`.

        The :class:`!LogicArray` is treated as an arbitrary-length vector of bits
        with the left-most bit being the most significant bit in the integer value.
        The bit vector is then interpreted as an integer using two's complement representation.

        :getter:
            Return the :class:`!LogicArray`'s value as a signed integer.

            .. deprecated:: 2.0
                Use :meth:`logic_array.to_signed() <cocotb.types.LogicArray.to_signed>` instead.

        :setter:
            Set the :class:`!LogicArray`'s value using a signed integer.

            .. deprecated:: 2.0
                Use ``logic_array[:] = LogicArray.from_signed(value, len(logic_array))`` instead.
        """
        return self.to_signed()

    @signed_integer.setter
    @deprecated(
        "`logic_array.signed_integer = value` setter is deprecated. "
        "Use `logic_array[:] = LogicArray.from_signed(value, len(logic_array))` instead."
    )
    def signed_integer(self, value: int) -> None:
        self[:] = LogicArray.from_signed(value, len(self))

    @property
    @deprecated(
        "`logic_array.buff` getter is deprecated. "
        'Use `logic_array.to_bytes(byteorder="big")` instead.'
    )
    def buff(self) -> bytes:
        """The :class:`!LogicArray`'s value as :class:`bytes`.

        The object is first converted to an :class:`int` as in :meth:`to_unsigned`.
        Then the object is converted to :class:`bytes` by converting the resulting integer value as in :meth:`int.to_bytes`.
        This assumes big-endian byte order and the minimal number of bytes necessary to hold any value of the current object.

        :getter:
            Return the :class:`!LogicArray`'s value as :class:`bytes`.

            .. deprecated:: 2.0
                Use :meth:`logic_array.to_bytes(byteorder="big") <cocotb.types.LogicArray.to_bytes>` instead.

        :setter:
            Set the :class:`!LogicArray`'s value using :class:`bytes`.

            .. deprecated:: 2.0
                Use ``logic_array[:] = LogicArray.from_bytes(value, len(logic_array), byteorder="big")`` instead.
        """
        return self.to_bytes(byteorder="big")

    @buff.setter
    @deprecated(
        "`logic_array.buff = value` setter is deprecated. "
        'Use `logic_array[:] = LogicArray.from_bytes(value, len(logic_array), byteorder="big")` instead.'
    )
    def buff(self, value: bytes) -> None:
        self[:] = LogicArray.from_bytes(value, len(self), byteorder="big")

    def to_unsigned(self) -> int:
        """Convert the value to an integer by interpreting it using unsigned representation.

        The :class:`!LogicArray` is treated as an arbitrary-length vector of bits
        with the left-most bit being the most significant bit in the integer value.
        The bit vector is then interpreted as an integer using unsigned representation.

        Returns:
            An integer equivalent to the value by interpreting it using unsigned representation.
        """
        if len(self) == 0:
            raise ValueError("Cannot convert null vector to integer")
        return self._get_int()

    def to_signed(self) -> int:
        """Convert the value to an integer by interpreting it using two's complement representation.

        The :class:`!LogicArray` is treated as an arbitrary-length vector of bits
        with the left-most bit being the most significant bit in the integer value.
        The bit vector is then interpreted as an integer using two's complement representation.

        Returns:
            An integer equivalent to the value by interpreting it using two's complement representation.
        """
        if len(self) == 0:
            raise ValueError("Cannot convert null vector to integer")
        value = self._get_int()
        limit = 1 << (len(self) - 1)
        if value >= limit:
            value -= 2 * limit
        return value

    def to_bytes(
        self,
        *,
        byteorder: ByteOrder,
    ) -> bytes:
        """Convert the value to bytes.

        The :class:`!LogicArray` is converted to an unsigned integer as in :meth:`to_unsigned`,
        then is converted to :class:`bytes` using *byteorder*-endian representation
        with the minimum number of bytes which can store all the bits in the original :class:`!LogicArray`.

        Args:
            byteorder: The endianness used to construct the intermediate integer, either ``"big"`` or ``"little"``.

        Returns:
            :class:`bytes` equivalent to the value.
        """
        return self.to_unsigned().to_bytes(ceil(len(self) / 8), byteorder=byteorder)

    @overload
    def __getitem__(self, item: int) -> Logic: ...

    @overload
    def __getitem__(self, item: slice) -> "LogicArray": ...

    def __getitem__(self, item: Union[int, slice]) -> Union[Logic, "LogicArray"]:
        array = self._get_array()
        if isinstance(item, int):
            if self._warn_indexing:
                warnings.warn(
                    f"Update index {item} to {self.range[item]}",
                    IndexingChangedWarning,
                    stacklevel=2,
                )
            idx = self._translate_index(item)
            return array[idx]
        elif isinstance(item, slice):
            if self._warn_indexing:
                start = item.start if item.start is not None else 0
                stop = item.stop if item.stop is not None else len(self) - 1
                warnings.warn(
                    f"Update slice {start}:{stop} to {self.range[start]}:{self.range[stop]}",
                    IndexingChangedWarning,
                    stacklevel=2,
                )
            start = item.start if item.start is not None else self.left
            stop = item.stop if item.stop is not None else self.right
            if item.step is not None:
                raise IndexError("do not specify step")
            start_i = self._translate_index(start)
            stop_i = self._translate_index(stop)
            if start_i > stop_i:
                raise IndexError(
                    f"slice [{start}:{stop}] direction does not match array direction [{self.left}:{self.right}]"
                )
            value = array[start_i : stop_i + 1]
            range = Range(start, self.direction, stop)
            return LogicArray(value=value, range=range)
        raise TypeError(f"indexes must be ints or slices, not {type(item).__name__}")

    @overload
    def __setitem__(self, item: int, value: LogicConstructibleT) -> None: ...

    @overload
    def __setitem__(
        self, item: slice, value: Union[str, Iterable[LogicConstructibleT], int]
    ) -> None: ...

    def __setitem__(
        self,
        item: Union[int, slice],
        value: Union[LogicConstructibleT, Iterable[LogicConstructibleT]],
    ) -> None:
        array = self._get_array()
        # invalid other impls
        self._value_as_str = None
        self._value_as_int = None
        if isinstance(item, int):
            idx = self._translate_index(item)
            array[idx] = Logic(cast("LogicConstructibleT", value))
        elif isinstance(item, slice):
            start = item.start if item.start is not None else self.left
            stop = item.stop if item.stop is not None else self.right
            if item.step is not None:
                raise IndexError("do not specify step")
            start_i = self._translate_index(start)
            stop_i = self._translate_index(stop)
            if start_i > stop_i:
                raise IndexError(
                    f"slice [{start}:{stop}] direction does not match array direction [{self.left}:{self.right}]"
                )
            value = cast("str | int | Iterable[LogicConstructibleT]", value)
            value_as_logics = LogicArray(value, stop_i - start_i + 1)
            array[start_i : stop_i + 1] = value_as_logics
        else:
            raise TypeError(
                f"indexes must be ints or slices, not {type(item).__name__}"
            )

    def _translate_index(self, item: int) -> int:
        try:
            return self._range.index(item)
        except ValueError:
            raise IndexError(f"index {item} out of range") from None

    def __repr__(self) -> str:
        return f"{type(self).__qualname__}({str(self)!r}, {self.range!r})"

    def __str__(self) -> str:
        return self._get_str()

    def __int__(self) -> int:
        return self.to_unsigned()

    def __index__(self) -> int:
        return int(self)

    def __and__(self, other: "LogicArray") -> "LogicArray":
        if not isinstance(other, LogicArray):
            return NotImplemented
        if len(self) != len(other):
            raise ValueError(
                f"cannot perform bitwise & "
                f"between {type(self).__qualname__} of length {len(self)} "
                f"and {type(other).__qualname__} of length {len(other)}"
            )
        return LogicArray(a & b for a, b in zip(self, other))

    def __or__(self, other: "LogicArray") -> "LogicArray":
        if not isinstance(other, LogicArray):
            return NotImplemented
        if len(self) != len(other):
            raise ValueError(
                f"cannot perform bitwise | "
                f"between {type(self).__qualname__} of length {len(self)} "
                f"and {type(other).__qualname__} of length {len(other)}"
            )
        return LogicArray(a | b for a, b in zip(self, other))

    def __xor__(self, other: "LogicArray") -> "LogicArray":
        if not isinstance(other, LogicArray):
            return NotImplemented
        if len(self) != len(other):
            raise ValueError(
                f"cannot perform bitwise ^ "
                f"between {type(self).__qualname__} of length {len(self)} "
                f"and {type(other).__qualname__} of length {len(other)}"
            )
        return LogicArray(a ^ b for a, b in zip(self, other))

    def __invert__(self) -> "LogicArray":
        return LogicArray(~v for v in self)

    if RESOLVE_X is None:

        def __bool__(self) -> bool:
            if len(self) == 0:
                return False
            return bool(int(self))

    else:

        def __bool__(self) -> bool:
            if len(self) == 0:
                return False
            return any(bool(bit) for bit in self)

    def resolve(self, resolver: ResolverLiteral) -> "LogicArray":
        """Resolves non-0/1 values to 0/1.

        The possible values of the *resolver* argument are:

        * ``"weak"``:
            Weak values are resolved to their strong-valued equivalents.

        * ``"zeros"``:
            ``L`` and ``H`` are resolved to ``0`` and ``1``, respectively.
            Remaining non-``0``/``1`` values are resolved to ``0``.

        * ``"ones"``:
            ``L`` and ``H`` are resolved to ``0`` and ``1``, respectively.
            Remaining non-``0``/``1`` values are resolved to ``1``.

        * ``"random"``:
            ``L`` and ``H`` are resolved to ``0`` and ``1``, respectively.
            Remaining non-``0``/``1`` values are randomly resolved to either ``0`` or ``1``.

        Args:
            resolver: How to resolve non-``0``/``1`` values. See possible values above.

        Returns:
            The resolved Logic.

        Raises:
            ValueError: Invalid *resolver* value.
            TypeError: Unsupported *value* type.
        """
        return LogicArray(get_str_resolver(resolver)(str(self)), self.range)

    def __copy__(self) -> "LogicArray":
        raise NotImplementedError("`copy.copy` on LogicArray is not supported")

    def __deepcopy__(self, memo: Dict[int, Any]) -> "LogicArray":
        res = LogicArray.__new__(LogicArray)
        res._value_as_array = copy.deepcopy(self._value_as_array, memo=memo)
        res._value_as_int = self._value_as_int
        res._value_as_str = self._value_as_str
        res._range = copy.deepcopy(self._range, memo=memo)
        res._warn_indexing = self._warn_indexing
        return res



# Copyright cocotb contributors
# Licensed under the Revised BSD License, see LICENSE for details.
# SPDX-License-Identifier: BSD-3-Clause
import collections.abc
import inspect
import logging
import traceback
from asyncio import CancelledError, InvalidStateError
from bdb import BdbQuit
from enum import auto
from types import SimpleNamespace
from typing import (
    TYPE_CHECKING,
    Callable,
    Coroutine,
    Generator,
    Generic,
    List,
    Optional,
    TypeVar,
    Union,
    cast,
)

import cocotb
from cocotb._base_triggers import Trigger
from cocotb._bridge import bridge, resume
from cocotb._deprecation import deprecated
from cocotb._outcomes import Error, Outcome, Value
from cocotb._py_compat import Self, cached_property
from cocotb._utils import DocEnum, extract_coro_stack, remove_traceback_frames

if TYPE_CHECKING:
    from types import CoroutineType


__all__ = (
    "Join",
    "Task",
    "TaskComplete",
    "bridge",
    "current_task",
    "resume",
)

# Set __module__ on re-exports
bridge.__module__ = __name__
resume.__module__ = __name__

#: Task result type
ResultType = TypeVar("ResultType")


class _TaskState(DocEnum):
    """State of a Task."""

    UNSTARTED = (auto(), "Task created, but never run and not scheduled")
    SCHEDULED = (auto(), "Task queued to run soon")
    PENDING = (auto(), "Task waiting for Trigger to fire")
    RUNNING = (auto(), "Task is currently running")
    FINISHED = (auto(), "Task has finished with a value or Exception")
    CANCELLED = (auto(), "Task was cancelled before it finished")


class Task(Generic[ResultType]):
    """Concurrently executing task.

    This class is not intended for users to directly instantiate.
    Use :func:`cocotb.create_task` to create a Task object
    or :func:`cocotb.start_soon` to create a Task and schedule it to run.

    .. versionchanged:: 1.8
        Moved to the ``cocotb.task`` module.

    .. versionchanged:: 2.0
        The ``retval``, ``_finished``, and ``__bool__`` methods were removed.
        Use :meth:`result`, :meth:`done`, and :meth:`done` methods instead, respectively.
    """

    _id_count = 0  # used by the scheduler for debug

    def __init__(
        self, inst: Coroutine[Trigger, None, ResultType], *, name: Optional[str] = None
    ) -> None:
        self._native_coroutine: bool
        if inspect.iscoroutinefunction(inst):
            raise TypeError(
                f"Coroutine function {inst} should be called prior to being scheduled."
            )
        elif inspect.isasyncgen(inst):
            raise TypeError(
                f"{inst.__qualname__} is an async generator, not a coroutine. "
                "You likely used the yield keyword instead of await."
            )
        elif inspect.iscoroutine(inst):
            self._native_coroutine = True
        elif isinstance(inst, collections.abc.Coroutine):
            self._native_coroutine = False
        else:
            raise TypeError(f"{inst} isn't a valid coroutine!")

        self._coro = inst
        self._state: _TaskState = _TaskState.UNSTARTED
        self._outcome: Union[Outcome[ResultType], None] = None
        self._trigger: Union[Trigger, None] = None
        self._done_callbacks: List[Callable[[Task[ResultType]], None]] = []
        self._cancelled_msg: Union[str, None] = None
        self._must_cancel: bool = False
        self._locals = SimpleNamespace()

        self._task_id = self._id_count
        type(self)._id_count += 1
        self._name = f"Task {self._task_id}" if name is None else name

    @property
    def locals(self) -> SimpleNamespace:
        '''Task-local variables.

        A modifiable namespace where any user-defined variables can be added.
        Use :func:`~cocotb.task.current_task` to access the current Task's locals.

        .. code-block:: python

            def get_rng() -> random.Random:
                """Get the random number generator for the current Task."""
                try:
                    return current_task().locals.rng
                except AttributeError:
                    rng = random.Random()
                    current_task().locals.rng = rng
                    return rng


            # Use Task-local RNG to drive a signal
            rng = get_rng()
            for _ in range(10):
                await drive(rng.randint(0, 100))

        .. versionadded:: 2.0
        '''
        return self._locals

    def get_name(self) -> str:
        """Return the name of the :class:`!Task`.

        If not set using :meth:`set_name` or passed during construction,
        a reasonable default name is generated.
        """
        return self._name

    def set_name(self, value: object) -> None:
        """Set the name of the :class:`!Task`.

        Args:
            value: Any object which can be converted to a :class:`str` to use as the name.
        """
        self._name = str(value)

    @cached_property
    def _cancelled_error(self) -> CancelledError:
        if self._cancelled_msg is None:
            return CancelledError()
        else:
            return CancelledError(self._cancelled_msg)

    @cached_property
    def _log(self) -> logging.Logger:
        coro_name: str
        if self._native_coroutine:
            coro_name = self._coro.__qualname__
        else:
            coro_name = type(self._coro).__qualname__
        return logging.getLogger(f"cocotb.{self._name}.{coro_name}")

    def __str__(self) -> str:
        # TODO Do we really need this?
        return f"<{self._name}>"

    def _get_coro_stack(self) -> traceback.StackSummary:
        """Get the coroutine callstack of this Task.

        Assumes :attr:`_coro` is a native Python coroutine object.

        Raises:
            TypeError: If :attr:`_coro` is not a native Python coroutine object.
        """
        if not self._native_coroutine:
            raise TypeError(
                "Task._get_coro_stack() can only be called on native Python coroutines."
            )

        coro_stack = extract_coro_stack(
            cast("CoroutineType[Trigger, None, ResultType]", self._coro)
        )

        # Remove Trigger.__await__() from the stack, as it's not really useful
        if len(coro_stack) > 0 and coro_stack[-1].name == "__await__":
            coro_stack.pop()

        return coro_stack

    def __repr__(self) -> str:
        if self._native_coroutine:
            coro_stack = self._get_coro_stack()
            try:
                coro_name = coro_stack[-1].name
            # coro_stack may be empty if:
            # - exhausted generator
            # - finished coroutine
            except IndexError:
                try:
                    coro_name = self._coro.__name__
                except AttributeError:
                    coro_name = type(self._coro).__name__
        else:
            coro_name = type(self._coro).__name__

        if self._state is _TaskState.RUNNING:
            return f"<{self._name} running coro={coro_name}()>"
        elif self._state is _TaskState.FINISHED:
            return f"<{self._name} finished coro={coro_name}() outcome={self._outcome}>"
        elif self._state is _TaskState.PENDING:
            return f"<{self._name} pending coro={coro_name}() trigger={self._trigger}>"
        elif self._state is _TaskState.SCHEDULED:
            return f"<{self._name} scheduled coro={coro_name}()>"
        elif self._state is _TaskState.UNSTARTED:
            return f"<{self._name} created coro={coro_name}()>"
        elif self._state is _TaskState.CANCELLED:
            return f"<{self._name} cancelled coro={coro_name} with={self._cancelled_error} outcome={self._outcome}"
        else:
            raise RuntimeError("Task in unknown state")

    def _set_outcome(
        self, result: Outcome[ResultType], state: _TaskState = _TaskState.FINISHED
    ) -> None:
        self._outcome = result
        self._state = state

        # Run done callbacks.
        for callback in self._done_callbacks:
            callback(self)

        # Wake up waiting Tasks.
        cocotb._scheduler_inst._react(self.complete)
        cocotb._scheduler_inst._react(self._join)

    def _advance(self, exc: Union[BaseException, None]) -> Union[Trigger, None]:
        """Resume execution of the Task.

        Runs until the coroutine ends, raises, or yields a Trigger.
        Can optionally throw an Exception into the Task.

        Args:
            exc: :exc:`BaseException` to throw into the coroutine or nothing.

        Returns:
            The object yielded from the coroutine or ``None`` if coroutine finished.
        """
        self._state = _TaskState.RUNNING

        if self._must_cancel:
            exc = self._cancelled_error

        try:
            if exc is None:
                trigger = self._coro.send(None)
            else:
                trigger = self._coro.throw(exc)
        except StopIteration as e:
            outcome = Value(e.value)
            if self._must_cancel:
                self._set_outcome(
                    Error(
                        RuntimeError(
                            "Task was cancelled, but exited normally. Did you forget to re-raise the CancelledError?"
                        )
                    )
                )
            else:
                self._set_outcome(outcome)
            return None
        except (KeyboardInterrupt, SystemExit, BdbQuit) as e:
            # Allow these to bubble up to the execution root to fail the sim immediately.
            # This follows asyncio's behavior.
            self._set_outcome(Error(remove_traceback_frames(e, ["_advance"])))
            raise
        except CancelledError as e:
            self._set_outcome(
                Error(remove_traceback_frames(e, ["_advance"])), _TaskState.CANCELLED
            )
            return None
        except BaseException as e:
            self._set_outcome(Error(remove_traceback_frames(e, ["_advance"])))
            return None
        else:
            if self._must_cancel:
                self._set_outcome(
                    Error(
                        RuntimeError(
                            "Task was cancelled, but continued running. Did you forget to re-raise the CancelledError?"
                        )
                    )
                )
                return None
            else:
                return trigger

    def _schedule_resume(self, exc: Optional[BaseException] = None) -> None:
        cocotb._scheduler_inst._unschedule(self)
        cocotb._scheduler_inst._schedule_task_internal(self, exc)

    @deprecated("`task.kill()` is deprecated in favor of `task.cancel()`")
    def kill(self) -> None:
        """Kill a coroutine."""

        if self._state in (_TaskState.PENDING, _TaskState.SCHEDULED):
            # Unschedule if scheduled and unprime triggers if pending.
            cocotb._scheduler_inst._unschedule(self)
        elif self._state is _TaskState.UNSTARTED:
            # Don't need to unschedule.
            pass
        elif self._state in (_TaskState.FINISHED, _TaskState.CANCELLED):
            # Do nothing if already done.
            return
        else:
            raise RuntimeError("Can't kill currently running Task")

        # Close native coroutines if they were never resumed to prevent ResourceWarnings.
        if (
            inspect.iscoroutine(self._coro)
            and inspect.getcoroutinestate(self._coro) == "CORO_CREATED"
        ):
            self._coro.close()

        self._set_outcome(Value(None))  # type: ignore  # `kill()` sets the result to None regardless of the ResultType

    @cached_property
    def complete(self) -> "TaskComplete[ResultType]":
        r"""Trigger which fires when the Task completes.

        Unlike :meth:`join`, this Trigger does not return the result of the Task when :keyword:`await`\ ed.

        .. code-block:: python

            async def coro_inner():
                await Timer(1, unit="ns")
                raise ValueError("Oops")


            task = cocotb.start_soon(coro_inner())
            await task.complete  # no exception raised here
            assert task.exception() == ValueError("Oops")
        """
        return TaskComplete._make(self)

    @deprecated(
        "Using `task` directly is preferred to `task.join()` in all situations where the latter could be used."
    )
    def join(self) -> "Join[ResultType]":
        r"""Block until the Task completes and return the result.

        Equivalent to calling :class:`Join(self) <cocotb.task.Join>`.

        .. code-block:: python

            async def coro_inner():
                await Timer(1, unit="ns")
                return "Hello world"


            task = cocotb.start_soon(coro_inner())
            result = await task.join()
            assert result == "Hello world"

        Returns:
            Object that can be :keyword:`await`\ ed or passed into :class:`~cocotb.triggers.First` or :class:`~cocotb.triggers.Combine`;
            the result of which will be the result of the Task.

        .. deprecated:: 2.0
            Using ``task`` directly is preferred to ``task.join()`` in all situations where the latter could be used.
        """
        return self._join

    @cached_property
    def _join(self) -> "Join[ResultType]":
        return Join._make(self)

    def cancel(self, msg: Optional[str] = None) -> bool:
        """Cancel a Task's further execution.

        When a Task is cancelled, a :exc:`asyncio.CancelledError` is thrown into the Task.

        Returns: ``True`` if the Task was cancelled; ``False`` otherwise.
        """
        if self._state in {_TaskState.PENDING, _TaskState.SCHEDULED}:
            self._schedule_resume()
        elif self._state in (_TaskState.UNSTARTED, _TaskState.RUNNING):
            # (Re)schedule to throw CancelledError
            cocotb._scheduler_inst._schedule_task_internal(self)
        else:
            # Already finished or cancelled
            return False

        self._cancelled_msg = msg
        self._must_cancel = True
        return True

    def _cancel_now(self, msg: Optional[str] = None) -> bool:
        """Like cancel(), but throws CancelledError into the Task and puts it into a "done" state immediately.

        Not safe to be called from a running Task.
        Only from done callbacks or scheduler or Task internals.
        """
        if self.done():
            return False

        self._cancelled_msg = msg
        self._must_cancel = True

        if self._state is _TaskState.UNSTARTED:
            # Must fail immediately as we can't start a coroutine with an exception.
            self._set_outcome(Error(self._cancelled_error), _TaskState.CANCELLED)
        else:
            # Unprime and unschedule the Task so it's out of the scheduler.
            cocotb._scheduler_inst._unschedule(self)
            # Force CancelledError to be thrown immediately.
            self._advance(None)

        return True

    def cancelled(self) -> bool:
        """Return ``True`` if the Task was cancelled."""
        return self._state is _TaskState.CANCELLED

    def done(self) -> bool:
        """Return ``True`` if the Task has finished executing."""
        return self._state in (_TaskState.FINISHED, _TaskState.CANCELLED)

    def result(self) -> ResultType:
        """Return the result of the Task.

        If the Task ran to completion, the result is returned.
        If the Task failed with an exception, the exception is re-raised.
        If the Task was cancelled, the :exc:`asyncio.CancelledError` is re-raised.
        If the coroutine is not yet complete, an :exc:`asyncio.InvalidStateError` is raised.
        """
        if self._state is _TaskState.CANCELLED:
            raise self._cancelled_error
        elif self._state is _TaskState.FINISHED:
            return cast("Outcome[ResultType]", self._outcome).get()
        else:
            raise InvalidStateError("result is not yet available")

    def exception(self) -> Optional[BaseException]:
        """Return the exception of the Task.

        If the Task ran to completion, ``None`` is returned.
        If the Task failed with an exception, the exception is returned.
        If the Task was cancelled, the :exc:`asyncio.CancelledError` is re-raised.
        If the coroutine is not yet complete, an :exc:`asyncio.InvalidStateError` is raised.
        """
        if self._state is _TaskState.CANCELLED:
            raise self._cancelled_error
        elif self._state is _TaskState.FINISHED:
            if isinstance(self._outcome, Error):
                return self._outcome.error
            else:
                return None
        else:
            raise InvalidStateError("result is not yet available")

    def _add_done_callback(
        self, callback: Callable[["Task[ResultType]"], None]
    ) -> None:
        """Add *callback* to the list of callbacks to be run once the Task becomes "done".

        Args:
            callback: The callback to run once "done".

        .. note::
            If the task is already done, calling this function will call the callback immediately.
        """
        if self.done():
            callback(self)
        self._done_callbacks.append(callback)

    def __await__(self) -> Generator[Trigger, None, ResultType]:
        if self._state is _TaskState.UNSTARTED:
            cocotb._scheduler_inst._schedule_task_internal(self)
            yield self.complete
        elif not self.done():
            yield self.complete
        return self.result()


def current_task() -> Task[object]:
    """Return the currently running Task.

    Raises:
        RuntimeError: If no Task is running.

    .. versionadded:: 2.0
    """
    task = cocotb._scheduler_inst._current_task
    if task is None:
        raise RuntimeError("No Task is currently running")
    return task


class TaskComplete(Trigger, Generic[ResultType]):
    r"""Fires when a :class:`~cocotb.task.Task` completes.

    Unlike :class:`~cocotb.task.Join`, this Trigger does not return the result of the Task when :keyword:`await`\ ed.
    See :attr:`.Task.complete` for more information.

    .. warning::
        This class cannot be instantiated in the normal way.
        You must use :attr:`.Task.complete`.

    .. versionadded:: 2.0
    """

    _task: Task[ResultType]

    def __new__(cls, task: Task[ResultType]) -> "TaskComplete[ResultType]":
        raise NotImplementedError(
            "TaskComplete cannot be instantiated in this way. Use the `task.complete` attribute."
        )

    @classmethod
    def _make(cls, task: Task[ResultType]) -> "Self":
        self = super().__new__(cls)
        super().__init__(self)
        self._task = task
        return self

    def _prime(self, callback: Callable[["Self"], None]) -> None:
        if self._task.done():
            callback(self)
        else:
            super()._prime(callback)

    def __repr__(self) -> str:
        return f"{type(self).__qualname__}({self._task!s})"

    @property
    def task(self) -> Task[ResultType]:
        """The :class:`.Task` associated with this completion event."""
        return self._task


class Join(TaskComplete[ResultType]):
    r"""Fires when a :class:`~cocotb.task.Task` completes and returns the Task's result.

    Equivalent to calling :meth:`task.join() <cocotb.task.Task.join>`.

    .. code-block:: python

        async def coro_inner():
            await Timer(1, unit="ns")
            return "Hello world"


        task = cocotb.start_soon(coro_inner())
        result = await Join(task)
        assert result == "Hello world"

    Args:
        task: The Task upon which to wait for completion.

    Returns:
        Object that can be :keyword:`await`\ ed or passed into :class:`~cocotb.triggers.First` or :class:`~cocotb.triggers.Combine`;
        the result of which will be the result of the Task.

    .. deprecated:: 2.0
        Using ``task`` directly is preferred to ``Join(task)`` in all situations where the latter could be used.
    """

    @deprecated(
        "Using `task` directly is preferred to `Join(task)` in all situations where the latter could be used."
    )
    def __new__(cls, task: Task[ResultType]) -> "Join[ResultType]":
        return task._join

    def __init__(self, task: Task[ResultType]) -> None:
        pass

    def __await__(self) -> Generator["Self", None, ResultType]:  # type: ignore[override]
        yield self
        return self._task.result()



# Copyright cocotb contributors
# Copyright (c) 2013 Potential Ventures Ltd
# Copyright (c) 2013 SolarFlare Communications Inc
# Licensed under the Revised BSD License, see LICENSE for details.
# SPDX-License-Identifier: BSD-3-Clause
import sys
from typing import TYPE_CHECKING, Dict, List, Union

from cocotb._decorators import (
    parametrize,
    test,
)
from cocotb._test import create_task, start, start_soon
from cocotb._test_functions import pass_test

from ._version import __version__ as _version

if TYPE_CHECKING:
    from logging import Logger
    from types import SimpleNamespace

    from cocotb._scheduler import Scheduler
    from cocotb.handle import SimHandleBase
    from cocotb.regression import RegressionManager

__all__ = (
    "RANDOM_SEED",
    "SIM_NAME",
    "SIM_VERSION",
    "__version__",
    "argv",
    "create_task",
    "is_simulation",
    "log",
    "packages",
    "parametrize",
    "pass_test",
    "plusargs",
    "start",
    "start_soon",
    "test",
    "top",
)

# Set __module__ on re-exports
test.__module__ = __name__
start_soon.__module__ = __name__
start.__module__ = __name__
create_task.__module__ = __name__
parametrize.__module__ = __name__
pass_test.__module__ = __name__


__version__: str = _version
"""The version of cocotb."""


log: "Logger"
"""An easily accessible :class:`~logging.Logger` for the user.

This logger defaults to the :data:`logging.INFO` log level.

.. versionchanged:: 2.0
    This was previously the ``"cocotb"`` Logger.
    It is now a Logger under the ``"test"`` namespace.
"""

_scheduler_inst: "Scheduler"
"""The global scheduler instance."""

_regression_manager: "RegressionManager"
"""The global regression manager instance."""

argv: List[str]
"""The argument list as seen by the simulator."""

plusargs: Dict[str, Union[bool, str]]
"""A dictionary of "plusargs" handed to the simulation.

See :envvar:`COCOTB_PLUSARGS` for details.
"""

packages: "SimpleNamespace"
"""A :class:`python:types.SimpleNamespace` of package handles.

This will be populated with handles at test time if packages can be discovered
via the GPI.

.. versionadded:: 2.0
"""

SIM_NAME: str
"""The product information of the running simulator."""

SIM_VERSION: str
"""The version of the running simulator."""

RANDOM_SEED: int
"""The last value used to seed the global PRNG.

During test collection, this is set to the value provided by :envvar:`COCOTB_RANDOM_SEED`,
if given, or a random value based on the state of the operating system.

During test run, this is set to a new value computed by combining the value used during test collection,
with the full test name (e.g. ``my_test_module.my_test``).
"""

top: "SimHandleBase"
r"""
A handle to the :envvar:`COCOTB_TOPLEVEL` entity/module.

This is equivalent to the :term:`DUT` parameter given to cocotb tests, so it can be used wherever that variable can be used.
It is particularly useful for extracting information about the :term:`DUT` in module-level class and function definitions;
and in parameters to :class:`.TestFactory`\ s.
"""

is_simulation: bool = False
"""``True`` if cocotb was loaded in a simulation."""


if sys.version_info < (3, 9):
    import warnings

    warnings.warn(
        "Support for Python versions < 3.9 will be dropped in version 2.1",
        FutureWarning,
    )



# Copyright cocotb contributors
# Licensed under the Revised BSD License, see LICENSE for details.
# SPDX-License-Identifier: BSD-3-Clause
"""
Backports and compatibility shims for newer python features.

These are for internal use - users should use a third party library like `six`
if they want to use these shims in their own code
"""

import sys
from abc import ABC
from contextlib import AbstractContextManager
from typing import TypeVar, Union, overload

__all__ = (
    "AbstractContextManager",
    "Final",
    "Literal",
    "ParamSpec",
    "Protocol",
    "Self",
    "StrEnum",
    "TypeAlias",
    "cached_property",
    "insertion_ordered_dict",
    "nullcontext",
)

T = TypeVar("T")

if sys.version_info < (3, 9):
    from typing import Generic

    T_co = TypeVar("T_co", covariant=True)
    ExitT_co = TypeVar("ExitT_co", covariant=True)

    class AbstractContextManager(AbstractContextManager, Generic[T_co, ExitT_co]): ...


# backport of Python 3.7's contextlib.nullcontext
class nullcontext(AbstractContextManager[T, None]):
    """Context manager that does no additional processing.
    Used as a stand-in for a normal context manager, when a particular
    block of code is only sometimes used with a normal context manager:

    cm = optional_cm if condition else nullcontext()
    with cm:
        # Perform operation, using optional_cm if condition is True
    """

    enter_result: T

    @overload
    def __init__(self: "nullcontext[None]", enter_result: None = None) -> None: ...

    @overload
    def __init__(self: "nullcontext[T]", enter_result: T) -> None: ...

    def __init__(
        self: "nullcontext[Union[T, None]]", enter_result: Union[T, None] = None
    ) -> None:
        self.enter_result = enter_result

    def __enter__(self) -> T:
        return self.enter_result

    def __exit__(self, *excinfo: object) -> None:
        pass


# On python 3.7 onwards, `dict` is guaranteed to preserve insertion order.
# Since `OrderedDict` is a little slower that `dict`, we prefer the latter
# when possible.
if sys.version_info[:2] >= (3, 7):  # noqa: UP036 | bug in ruff
    insertion_ordered_dict = dict
else:
    import collections

    insertion_ordered_dict = collections.OrderedDict


# simple, but less than optimal backport of Python 3.8's cached_property
if sys.version_info >= (3, 8):
    from functools import cached_property
else:
    from functools import update_wrapper

    class cached_property:
        def __init__(self, method):
            self._method = method
            update_wrapper(self, method)

        def __get__(self, instance, owner=None):
            if instance is None:
                return self
            res = self._method(instance)
            instance.__dict__[self._method.__name__] = res
            return res


if sys.version_info >= (3, 8):
    from typing import Final, Literal, Protocol
else:
    from typing import Any

    class FakeGetItemMetatype(type):
        def __getitem__(cls, a: object) -> Any:
            return Any

    class FakeGetItemType(metaclass=FakeGetItemMetatype): ...

    Final = FakeGetItemType
    Literal = FakeGetItemType
    Protocol = ABC


if sys.version_info >= (3, 10):
    from typing import ParamSpec, TypeAlias
else:
    from typing import Any

    TypeAlias = Any

    class FakeParamSpecType:
        def __init__(self, name: str) -> None: ...

        def kwargs(self) -> Any:
            return Any

        def args(self) -> Any:
            return Any

    ParamSpec = FakeParamSpecType


if sys.version_info >= (3, 11):
    from typing import Self
else:
    Self = ""


if sys.version_info >= (3, 11):
    from enum import StrEnum
else:
    from enum import Enum

    class StrEnum(str, Enum):
        def __str__(self) -> str:
            return self.value



# Copyright cocotb contributors
# Copyright (c) 2013, 2018 Potential Ventures Ltd
# Copyright (c) 2013 SolarFlare Communications Inc
# Licensed under the Revised BSD License, see LICENSE for details.
# SPDX-License-Identifier: BSD-3-Clause

"""All things relating to regression capabilities."""

import functools
import hashlib
import inspect
import logging
import os
import random
import re
import time
import warnings
from enum import auto
from importlib import import_module
from typing import TYPE_CHECKING, Any, Callable, Coroutine, List, Union

import cocotb
import cocotb._gpi_triggers
import cocotb.handle
from cocotb import logging as cocotb_logging
from cocotb import simulator
from cocotb._decorators import Parameterized, Test
from cocotb._extended_awaitables import with_timeout
from cocotb._gpi_triggers import GPITrigger, Timer
from cocotb._outcomes import Error, Outcome
from cocotb._test import RunningTest
from cocotb._test_factory import TestFactory
from cocotb._test_functions import Failed
from cocotb._utils import (
    DocEnum,
    remove_traceback_frames,
    safe_divide,
)
from cocotb._xunit_reporter import XUnitReporter, bin_xml_escape
from cocotb.logging import ANSI
from cocotb.simtime import get_sim_time
from cocotb.task import Task

if TYPE_CHECKING:
    from cocotb._base_triggers import Trigger

__all__ = (
    "Parameterized",
    "RegressionManager",
    "RegressionMode",
    "SimFailure",
    "Test",
    "TestFactory",
)

# Set __module__ on re-exports
Parameterized.__module__ = __name__
Test.__module__ = __name__
TestFactory.__module__ = __name__


class SimFailure(BaseException):
    """A Test failure due to simulator failure.

    .. caution::
        Not to be raised or caught within a test.
        Only used for marking expected failure with ``expect_error`` in :func:`cocotb.test`.
    """


_logger = logging.getLogger(__name__)


def _format_doc(docstring: Union[str, None]) -> str:
    if docstring is None:
        return ""
    else:
        brief = docstring.split("\n")[0]
        return f"\n    {brief}"


class RegressionMode(DocEnum):
    """The mode of the :class:`RegressionManager`."""

    REGRESSION = (
        auto(),
        """Tests are run if included. Skipped tests are skipped, expected failures and errors are respected.""",
    )

    TESTCASE = (
        auto(),
        """Like :attr:`REGRESSION`, but skipped tests are *not* skipped if included.""",
    )


class _TestResults:
    # TODO Replace with dataclass in Python 3.7+

    def __init__(
        self,
        test_fullname: str,
        passed: Union[None, bool],
        wall_time_s: float,
        sim_time_ns: float,
    ) -> None:
        self.test_fullname = test_fullname
        self.passed = passed
        self.wall_time_s = wall_time_s
        self.sim_time_ns = sim_time_ns

    @property
    def ratio(self) -> float:
        return safe_divide(self.sim_time_ns, self.wall_time_s)


class RegressionManager:
    """Object which manages tests.

    This object uses the builder pattern to build up a regression.
    Tests are added using :meth:`register_test` or :meth:`discover_tests`.
    Inclusion filters for tests can be added using :meth:`add_filters`.
    The "mode" of the regression can be controlled using :meth:`set_mode`.
    These methods can be called in any order any number of times before :meth:`start_regression` is called,
    and should not be called again after that.

    Once all the tests, filters, and regression behavior configuration is done,
    the user starts the regression with :meth:`start_regression`.
    This method must be called exactly once.

    Until the regression is started, :attr:`total_tests`, :attr:`count`, :attr:`passed`,
    :attr:`skipped`, and :attr:`failures` hold placeholder values.
    """

    COLOR_TEST = ANSI.BLUE_FG
    COLOR_PASSED = ANSI.GREEN_FG
    COLOR_SKIPPED = ANSI.YELLOW_FG
    COLOR_FAILED = ANSI.RED_FG

    _timer1 = Timer(1)

    def __init__(self) -> None:
        self._test: Test
        self._running_test: RunningTest
        self.log = _logger
        self._regression_start_time: float
        self._test_results: List[_TestResults] = []
        self.total_tests = 0
        """Total number of tests that will be run or skipped."""
        self.count = 0
        """The current test count."""
        self.passed = 0
        """The current number of passed tests."""
        self.skipped = 0
        """The current number of skipped tests."""
        self.failures = 0
        """The current number of failed tests."""
        self._tearing_down = False
        self._test_queue: List[Test] = []
        self._filters: List[re.Pattern[str]] = []
        self._mode = RegressionMode.REGRESSION
        self._included: List[bool]
        self._sim_failure: Union[Error[None], None] = None
        self._regression_seed = cocotb.RANDOM_SEED
        self._random_state: Any

        # Setup XUnit
        ###################

        results_filename = os.getenv("COCOTB_RESULTS_FILE", "results.xml")
        suite_name = os.getenv("COCOTB_RESULT_TESTSUITE", "all")
        package_name = os.getenv("COCOTB_RESULT_TESTPACKAGE", "all")

        self.xunit = XUnitReporter(filename=results_filename)
        self.xunit.add_testsuite(name=suite_name, package=package_name)
        self.xunit.add_property(name="random_seed", value=str(cocotb.RANDOM_SEED))

    def discover_tests(self, *modules: str) -> None:
        """Discover tests in files automatically.

        Should be called before :meth:`start_regression` is called.

        Args:
            modules: Each argument given is the name of a module where tests are found.
        """
        for module_name in modules:
            mod = import_module(module_name)

            found_test: bool = False
            for obj_name, obj in vars(mod).items():
                if isinstance(obj, Test):
                    found_test = True
                    self.register_test(obj)
                elif isinstance(obj, Parameterized):
                    found_test = True
                    generated_tests: bool = False
                    for test in obj.generate_tests():
                        generated_tests = True
                        self.register_test(test)
                    if not generated_tests:
                        warnings.warn(
                            f"Parametrize object generated no tests: {module_name}.{obj_name}",
                            stacklevel=2,
                        )

            if not found_test:
                warnings.warn(
                    f"No tests were discovered in module: {module_name}", stacklevel=2
                )

        # error if no tests were discovered
        if not self._test_queue:
            modules_str = ", ".join(repr(m) for m in modules)
            raise RuntimeError(f"No tests were discovered in any module: {modules_str}")

    def add_filters(self, *filters: str) -> None:
        """Add regular expressions to filter-in registered tests.

        Only those tests which match at least one of the given filters are included;
        the rest are excluded.

        Should be called before :meth:`start_regression` is called.

        Args:
            filters: Each argument given is a regex pattern for test names.
                A match *includes* the test.
        """
        for filter in filters:
            compiled_filter = re.compile(filter)
            self._filters.append(compiled_filter)

    def set_mode(self, mode: RegressionMode) -> None:
        """Set the regression mode.

        See :class:`RegressionMode` for more details on how each mode affects :class:`RegressionManager` behavior.
        Should be called before :meth:`start_regression` is called.

        Args:
            mode: The regression mode to set.
        """
        self._mode = mode

    def register_test(self, test: Test) -> None:
        """Register a test with the :class:`RegressionManager`.

        Should be called before :meth:`start_regression` is called.

        Args:
            test: The test object to register.
        """
        self.log.debug("Registered test %r", test.fullname)
        self._test_queue.append(test)

    @classmethod
    def setup_pytest_assertion_rewriting(cls) -> None:
        """Configure pytest to rewrite assertions for better failure messages.

        Must be called before all modules containing tests are imported.
        """
        try:
            import pytest  # noqa: PLC0415
        except ImportError:
            _logger.info(
                "pytest not found, install it to enable better AssertionError messages"
            )
            return
        try:
            # Install the assertion rewriting hook, which must be done before we
            # import the test modules.
            from _pytest.assertion import install_importhook  # noqa: PLC0415
            from _pytest.config import Config  # noqa: PLC0415

            python_files = os.getenv("COCOTB_REWRITE_ASSERTION_FILES", "*.py").strip()
            if not python_files:
                # Even running the hook causes exceptions in some cases, so if the user
                # selects nothing, don't install the hook at all.
                return

            pytest_conf = Config.fromdictargs(
                {}, ["--capture=no", "-o", f"python_files={python_files}"]
            )
            install_importhook(pytest_conf)
        except Exception:
            _logger.exception(
                "Configuring the assertion rewrite hook using pytest %s failed. "
                "Please file a bug report!",
                pytest.__version__,
            )

    def start_regression(self) -> None:
        """Start the regression."""

        # sort tests into stages
        self._test_queue.sort(key=lambda test: test.stage)

        # mark tests for running
        if self._filters:
            self._included = [False] * len(self._test_queue)
            for i, test in enumerate(self._test_queue):
                for filter in self._filters:
                    if filter.search(test.fullname):
                        self._included[i] = True
        else:
            self._included = [True] * len(self._test_queue)

        # compute counts
        self.count = 1
        self.total_tests = sum(self._included)
        if self.total_tests == 0:
            self.log.warning(
                "No tests left after filtering with: %s",
                ", ".join(f.pattern for f in self._filters),
            )

        # start write scheduler
        cocotb.handle._start_write_scheduler()

        # start test loop
        self._regression_start_time = time.time()
        self._first_test = True
        self._execute()

    def _execute(self) -> None:
        """Run the main regression loop.

        Used by :meth:`start_regression` and :meth:`_test_complete` to continue to the main test running loop,
        and by :meth:`_fail_regression` to shutdown the regression when a simulation failure occurs.
        """

        while self._test_queue:
            self._test = self._test_queue.pop(0)
            included = self._included.pop(0)

            # if the test is not included, record and continue
            if not included:
                self._record_test_excluded()
                continue

            # if the test is skipped, record and continue
            if self._test.skip and self._mode != RegressionMode.TESTCASE:
                self._record_test_skipped()
                continue

            # if the test should be run, but the simulator has failed, record and continue
            if self._sim_failure is not None:
                self._score_test(
                    self._sim_failure,
                    0,
                    0,
                )
                continue

            # initialize the test, if it fails, record and continue
            try:
                self._running_test = self._init_test()
            except Exception:
                self._record_test_init_failed()
                continue

            self._log_test_start()

            if self._first_test:
                self._first_test = False
                return self._schedule_next_test()
            else:
                return self._timer1._prime(self._schedule_next_test)

        return self._tear_down()

    def _init_test(self) -> RunningTest:
        # wrap test function in timeout
        func: Callable[..., Coroutine[Trigger, None, None]]
        timeout = self._test.timeout_time
        if timeout is not None:
            f = self._test.func

            @functools.wraps(f)
            async def func(*args: object, **kwargs: object) -> None:
                await with_timeout(f(*args, **kwargs), timeout, self._test.timeout_unit)
        else:
            func = self._test.func

        main_task = Task(func(cocotb.top), name=f"Test {self._test.name}")
        return RunningTest(self._test_complete, main_task)

    def _schedule_next_test(self, trigger: Union[GPITrigger, None] = None) -> None:
        if trigger is not None:
            # TODO move to Trigger object
            cocotb._gpi_triggers._current_gpi_trigger = trigger
            trigger._cleanup()

        # seed random number generator based on test module, name, and COCOTB_RANDOM_SEED
        hasher = hashlib.sha1()
        hasher.update(self._test.fullname.encode())
        test_seed = self._regression_seed + int(hasher.hexdigest(), 16)

        cocotb.RANDOM_SEED = test_seed
        self._random_state = random.getstate()
        random.seed(test_seed)

        self._start_sim_time = get_sim_time("ns")
        self._start_time = time.time()

        self._running_test.start()

    def _tear_down(self) -> None:
        """Called by :meth:`_execute` when there are no more tests to run to finalize the regression."""
        # prevent re-entering the tear down procedure
        if not self._tearing_down:
            self._tearing_down = True
        else:
            return

        assert not self._test_queue

        # stop the write scheduler
        cocotb.handle._stop_write_scheduler()

        # Write out final log messages
        self._log_test_summary()

        # Generate output reports
        self.xunit.write()

        # TODO refactor initialization and finalization into their own module
        # to prevent circular imports requiring local imports
        from cocotb._init import _shutdown_testbench  # noqa: PLC0415

        _shutdown_testbench()

        # Setup simulator finalization
        simulator.stop_simulator()

    def _test_complete(self) -> None:
        """Callback given to the test to be called when the test finished."""

        # compute wall time
        wall_time = time.time() - self._start_time
        sim_time_ns = get_sim_time("ns") - self._start_sim_time

        cocotb.RANDOM_SEED = self._regression_seed
        random.setstate(self._random_state)

        # Judge and record pass/fail.
        self._score_test(
            self._running_test.result(),
            wall_time,
            sim_time_ns,
        )

        # Run next test.
        return self._execute()

    def _score_test(
        self,
        outcome: Outcome[None],
        wall_time_s: float,
        sim_time_ns: float,
    ) -> None:
        test = self._test

        # score test
        passed: bool
        msg: Union[str, None]
        exc: Union[BaseException, None]
        try:
            outcome.get()
        except BaseException as e:
            passed, msg = False, None
            exc = remove_traceback_frames(e, ["_score_test", "get"])
        else:
            passed, msg, exc = True, None, None

        if passed:
            if test.expect_error:
                self._record_test_failed(
                    wall_time_s=wall_time_s,
                    sim_time_ns=sim_time_ns,
                    result=exc,
                    msg="passed but we expected an error",
                )
                passed = False

            elif test.expect_fail:
                self._record_test_failed(
                    wall_time_s=wall_time_s,
                    sim_time_ns=sim_time_ns,
                    result=exc,
                    msg="passed but we expected a failure",
                )
                passed = False

            else:
                self._record_test_passed(
                    wall_time_s=wall_time_s,
                    sim_time_ns=sim_time_ns,
                    result=None,
                    msg=msg,
                )

        elif test.expect_fail:
            if isinstance(exc, (AssertionError, Failed)):
                self._record_test_passed(
                    wall_time_s=wall_time_s,
                    sim_time_ns=sim_time_ns,
                    result=None,
                    msg="failed as expected",
                )

            else:
                self._record_test_failed(
                    wall_time_s=wall_time_s,
                    sim_time_ns=sim_time_ns,
                    result=exc,
                    msg="expected failure, but errored with unexpected type",
                )
                passed = False

        elif test.expect_error:
            if isinstance(exc, test.expect_error):
                self._record_test_passed(
                    wall_time_s=wall_time_s,
                    sim_time_ns=sim_time_ns,
                    result=None,
                    msg="errored as expected",
                )

            else:
                self._record_test_failed(
                    wall_time_s=wall_time_s,
                    sim_time_ns=sim_time_ns,
                    result=exc,
                    msg="errored with unexpected type",
                )
                passed = False

        else:
            self._record_test_failed(
                wall_time_s=wall_time_s,
                sim_time_ns=sim_time_ns,
                result=exc,
                msg=msg,
            )

    def _get_lineno(self, test: Test) -> int:
        try:
            return inspect.getsourcelines(test.func)[1]
        except OSError:
            return 1

    def _log_test_start(self) -> None:
        """Called by :meth:`_execute` to log that a test is starting."""
        hilight_start = "" if cocotb_logging.strip_ansi else self.COLOR_TEST
        hilight_end = "" if cocotb_logging.strip_ansi else ANSI.DEFAULT
        self.log.info(
            "%srunning%s %s (%d/%d)%s",
            hilight_start,
            hilight_end,
            self._test.fullname,
            self.count,
            self.total_tests,
            _format_doc(self._test.doc),
        )

    def _record_test_excluded(self) -> None:
        """Called by :meth:`_execute` when a test is excluded by filters."""

        # write out xunit results
        lineno = self._get_lineno(self._test)
        self.xunit.add_testcase(
            name=self._test.name,
            classname=self._test.module,
            file=inspect.getfile(self._test.func),
            lineno=repr(lineno),
            time=repr(0),
            sim_time_ns=repr(0),
            ratio_time=repr(0),
        )
        self.xunit.add_skipped()

        # do not log anything, nor save details for the summary

    def _record_test_skipped(self) -> None:
        """Called by :meth:`_execute` when a test is skipped."""

        # log test results
        hilight_start = "" if cocotb_logging.strip_ansi else self.COLOR_SKIPPED
        hilight_end = "" if cocotb_logging.strip_ansi else ANSI.DEFAULT
        self.log.info(
            "%sskipping%s %s (%d/%d)%s",
            hilight_start,
            hilight_end,
            self._test.fullname,
            self.count,
            self.total_tests,
            _format_doc(self._test.doc),
        )

        # write out xunit results
        lineno = self._get_lineno(self._test)
        self.xunit.add_testcase(
            name=self._test.name,
            classname=self._test.module,
            file=inspect.getfile(self._test.func),
            lineno=repr(lineno),
            time=repr(0),
            sim_time_ns=repr(0),
            ratio_time=repr(0),
        )
        self.xunit.add_skipped()

        # save details for summary
        self._test_results.append(
            _TestResults(
                test_fullname=self._test.fullname,
                passed=None,
                sim_time_ns=0,
                wall_time_s=0,
            )
        )

        # update running passed/failed/skipped counts
        self.skipped += 1
        self.count += 1

    def _record_test_init_failed(self) -> None:
        """Called by :meth:`_execute` when a test initialization fails."""

        # log test results
        hilight_start = "" if cocotb_logging.strip_ansi else self.COLOR_FAILED
        hilight_end = "" if cocotb_logging.strip_ansi else ANSI.DEFAULT
        self.log.exception(
            "%sFailed to initialize%s %s! (%d/%d)%s",
            hilight_start,
            hilight_end,
            self._test.fullname,
            self.count,
            self.total_tests,
            _format_doc(self._test.doc),
        )

        # write out xunit results
        lineno = self._get_lineno(self._test)
        self.xunit.add_testcase(
            name=self._test.name,
            classname=self._test.module,
            file=inspect.getfile(self._test.func),
            lineno=repr(lineno),
            time=repr(0),
            sim_time_ns=repr(0),
            ratio_time=repr(0),
        )
        self.xunit.add_failure(msg="Test initialization failed")

        # save details for summary
        self._test_results.append(
            _TestResults(
                test_fullname=self._test.fullname,
                passed=False,
                sim_time_ns=0,
                wall_time_s=0,
            )
        )

        # update running passed/failed/skipped counts
        self.failures += 1
        self.count += 1

    def _record_test_passed(
        self,
        wall_time_s: float,
        sim_time_ns: float,
        result: Union[Exception, None],
        msg: Union[str, None],
    ) -> None:
        start_hilight = "" if cocotb_logging.strip_ansi else self.COLOR_PASSED
        stop_hilight = "" if cocotb_logging.strip_ansi else ANSI.DEFAULT
        if msg is None:
            rest = ""
        else:
            rest = f": {msg}"
        if result is None:
            result_was = ""
        else:
            result_was = f" (result was {type(result).__qualname__})"
        self.log.info(
            "%s %spassed%s%s%s",
            self._test.fullname,
            start_hilight,
            stop_hilight,
            rest,
            result_was,
        )

        # write out xunit results
        ratio_time = safe_divide(sim_time_ns, wall_time_s)
        lineno = self._get_lineno(self._test)
        self.xunit.add_testcase(
            name=self._test.name,
            classname=self._test.module,
            file=inspect.getfile(self._test.func),
            lineno=repr(lineno),
            time=repr(wall_time_s),
            sim_time_ns=repr(sim_time_ns),
            ratio_time=repr(ratio_time),
        )

        # update running passed/failed/skipped counts
        self.passed += 1
        self.count += 1

        # save details for summary
        self._test_results.append(
            _TestResults(
                test_fullname=self._test.fullname,
                passed=True,
                sim_time_ns=sim_time_ns,
                wall_time_s=wall_time_s,
            )
        )

    def _record_test_failed(
        self,
        wall_time_s: float,
        sim_time_ns: float,
        result: Union[BaseException, None],
        msg: Union[str, None],
    ) -> None:
        start_hilight = "" if cocotb_logging.strip_ansi else self.COLOR_FAILED
        stop_hilight = "" if cocotb_logging.strip_ansi else ANSI.DEFAULT
        if msg is None:
            rest = ""
        else:
            rest = f": {msg}"
        self.log.warning(
            "%s%s %sfailed%s%s",
            stop_hilight,
            self._test.fullname,
            start_hilight,
            stop_hilight,
            rest,
        )

        # write out xunit results
        ratio_time = safe_divide(sim_time_ns, wall_time_s)
        lineno = self._get_lineno(self._test)
        self.xunit.add_testcase(
            name=self._test.name,
            classname=self._test.module,
            file=inspect.getfile(self._test.func),
            lineno=repr(lineno),
            time=repr(wall_time_s),
            sim_time_ns=repr(sim_time_ns),
            ratio_time=repr(ratio_time),
        )
        self.xunit.add_failure(
            error_type=type(result).__name__, error_msg=bin_xml_escape(result)
        )

        # update running passed/failed/skipped counts
        self.failures += 1
        self.count += 1

        # save details for summary
        self._test_results.append(
            _TestResults(
                test_fullname=self._test.fullname,
                passed=False,
                sim_time_ns=sim_time_ns,
                wall_time_s=wall_time_s,
            )
        )

    def _log_test_summary(self) -> None:
        """Called by :meth:`_tear_down` to log the test summary."""
        real_time = time.time() - self._regression_start_time
        sim_time_ns = get_sim_time("ns")
        ratio_time = safe_divide(sim_time_ns, real_time)

        if len(self._test_results) == 0:
            return

        TEST_FIELD = "TEST"
        RESULT_FIELD = "STATUS"
        SIM_FIELD = "SIM TIME (ns)"
        REAL_FIELD = "REAL TIME (s)"
        RATIO_FIELD = "RATIO (ns/s)"
        TOTAL_NAME = f"TESTS={self.total_tests} PASS={self.passed} FAIL={self.failures} SKIP={self.skipped}"

        TEST_FIELD_LEN = max(
            len(TEST_FIELD),
            len(TOTAL_NAME),
            len(max([x.test_fullname for x in self._test_results], key=len)),
        )
        RESULT_FIELD_LEN = len(RESULT_FIELD)
        SIM_FIELD_LEN = len(SIM_FIELD)
        REAL_FIELD_LEN = len(REAL_FIELD)
        RATIO_FIELD_LEN = len(RATIO_FIELD)

        header_dict = {
            "a": TEST_FIELD,
            "b": RESULT_FIELD,
            "c": SIM_FIELD,
            "d": REAL_FIELD,
            "e": RATIO_FIELD,
            "a_len": TEST_FIELD_LEN,
            "b_len": RESULT_FIELD_LEN,
            "c_len": SIM_FIELD_LEN,
            "d_len": REAL_FIELD_LEN,
            "e_len": RATIO_FIELD_LEN,
        }

        LINE_LEN = (
            3
            + TEST_FIELD_LEN
            + 2
            + RESULT_FIELD_LEN
            + 2
            + SIM_FIELD_LEN
            + 2
            + REAL_FIELD_LEN
            + 2
            + RATIO_FIELD_LEN
            + 3
        )

        LINE_SEP = "*" * LINE_LEN + "\n"

        summary = ""
        summary += LINE_SEP
        summary += "** {a:<{a_len}}  {b:^{b_len}}  {c:>{c_len}}  {d:>{d_len}}  {e:>{e_len}} **\n".format(
            **header_dict
        )
        summary += LINE_SEP

        test_line = "** {a:<{a_len}}  {start}{b:^{b_len}}{end}  {c:>{c_len}.2f}   {d:>{d_len}.2f}   {e:>{e_len}}  **\n"
        hilite: str
        lolite: str
        for result in self._test_results:
            if result.passed is None:
                ratio = "-.--"
                pass_fail_str = "SKIP"
                hilite = self.COLOR_SKIPPED
                lolite = ANSI.DEFAULT
            elif result.passed:
                ratio = format(result.ratio, "0.2f")
                pass_fail_str = "PASS"
                hilite = self.COLOR_PASSED
                lolite = ANSI.DEFAULT
            else:
                ratio = format(result.ratio, "0.2f")
                pass_fail_str = "FAIL"
                hilite = self.COLOR_FAILED
                lolite = ANSI.DEFAULT

            if cocotb_logging.strip_ansi:
                hilite = ""
                lolite = ""

            test_dict = {
                "a": result.test_fullname,
                "b": pass_fail_str,
                "c": result.sim_time_ns,
                "d": result.wall_time_s,
                "e": ratio,
                "a_len": TEST_FIELD_LEN,
                "b_len": RESULT_FIELD_LEN,
                "c_len": SIM_FIELD_LEN - 1,
                "d_len": REAL_FIELD_LEN - 1,
                "e_len": RATIO_FIELD_LEN - 1,
                "start": hilite,
                "end": lolite,
            }

            summary += test_line.format(**test_dict)

        summary += LINE_SEP

        summary += test_line.format(
            a=TOTAL_NAME,
            b="",
            c=sim_time_ns,
            d=real_time,
            e=format(ratio_time, "0.2f"),
            a_len=TEST_FIELD_LEN,
            b_len=RESULT_FIELD_LEN,
            c_len=SIM_FIELD_LEN - 1,
            d_len=REAL_FIELD_LEN - 1,
            e_len=RATIO_FIELD_LEN - 1,
            start="",
            end="",
        )

        summary += LINE_SEP

        self.log.info(summary)

    def _fail_simulation(self, msg: str) -> None:
        self._sim_failure = Error(SimFailure(msg))
        self._running_test.abort(self._sim_failure)
        cocotb._scheduler_inst._event_loop()



# Copyright cocotb contributors
# Licensed under the Revised BSD License, see LICENSE for details.
# SPDX-License-Identifier: BSD-3-Clause
import functools
import logging
import threading
from enum import IntEnum
from typing import (
    TYPE_CHECKING,
    Callable,
    Coroutine,
    Generic,
    TypeVar,
    Union,
)

import cocotb
from cocotb import debug
from cocotb._base_triggers import Event, Trigger
from cocotb._exceptions import InternalError
from cocotb._py_compat import ParamSpec

if TYPE_CHECKING:
    from cocotb._outcomes import Outcome

P = ParamSpec("P")

Result = TypeVar("Result")


def resume(
    func: "Callable[P, Coroutine[Trigger, None, Result]]",
) -> "Callable[P, Result]":
    """Converts a coroutine function into a blocking function.

    This allows a :term:`coroutine function` that awaits cocotb triggers to be
    called from a :term:`blocking function` converted by :func:`.bridge`.
    This completes the bridge through non-:keyword:`async` code.

    When a converted coroutine function is called the current function blocks until the
    converted function exits.

    Results of the converted function are returned from the function call.

    Args:
        func: The :term:`coroutine function` to convert into a :term:`blocking function`.

    Returns:
        *func* as a :term:`blocking function`.

    Raises:
        RuntimeError:
            If the function that is returned is subsequently called from a
            thread that was not started with :class:`.bridge`.

    .. versionchanged:: 2.0
        Renamed from ``function``.
        No longer implemented as a type.
        The ``log`` attribute is no longer available.
    """

    @functools.wraps(func)
    def wrapper(*args: "P.args", **kwargs: "P.kwargs") -> Result:
        return cocotb._scheduler_inst._queue_function(func(*args, **kwargs))

    return wrapper


def bridge(
    func: "Callable[P, Result]",
) -> "Callable[P, Coroutine[Trigger, None, Result]]":
    r"""Converts a blocking function into a coroutine function.

    This function converts a :term:`blocking function` into a :term:`coroutine function`
    with the expectation that the function being converted is intended to call a
    :func:`.resume` converted function. This creates a bridge through
    non-:keyword:`async` code for code wanting to eventually :keyword:`await` on cocotb
    triggers.

    When a converted function call is used in an :keyword:`await` statement, the current
    Task blocks until the converted function finishes.

    Results of the converted function are returned from the :keyword:`await` expression.

    .. note::
        Bridge threads *must* either finish or block on a :func:`.resume`
        converted function before control is given back to the simulator.
        This is done to prevent any code from executing in parallel with the simulation.

    Args:
        func: The :term:`blocking function` to convert into a :term:`coroutine function`.

    Returns:
        *func* as a :term:`coroutine function`.

    .. versionchanged:: 2.0
        Renamed from ``external``.
        No longer implemented as a type.
        The ``log`` attribute is no longer available.
    """

    @functools.wraps(func)
    def wrapper(
        *args: "P.args", **kwargs: "P.kwargs"
    ) -> Coroutine[Trigger, None, Result]:
        return cocotb._scheduler_inst._run_in_executor(func, *args, **kwargs)

    return wrapper


class external_state(IntEnum):
    INIT = 0
    RUNNING = 1
    PAUSED = 2
    EXITED = 3


class external_waiter(Generic[Result]):
    def __init__(self) -> None:
        self._outcome: Union[Outcome[Result], None] = None
        self.thread: threading.Thread
        self.event = Event()
        self.state = external_state.INIT
        self.cond = threading.Condition()
        self._log = logging.getLogger(f"cocotb.bridge.0x{id(self):x}")

    @property
    def result(self) -> Result:
        if self._outcome is None:
            raise InternalError("Got result of external before it finished")
        return self._outcome.get()

    def _propagate_state(self, new_state: external_state) -> None:
        with self.cond:
            if debug.debug:
                self._log.debug(
                    f"Changing state from {self.state} -> {new_state} from {threading.current_thread()}"
                )
            self.state = new_state
            self.cond.notify()

    def thread_done(self) -> None:
        if debug.debug:
            self._log.debug(f"Thread finished from {threading.current_thread()}")
        self._propagate_state(external_state.EXITED)

    def thread_suspend(self) -> None:
        self._propagate_state(external_state.PAUSED)

    def thread_start(self) -> None:
        if self.state > external_state.INIT:
            return

        if not self.thread.is_alive():
            self._propagate_state(external_state.RUNNING)
            self.thread.start()

    def thread_resume(self) -> None:
        self._propagate_state(external_state.RUNNING)

    def thread_wait(self) -> external_state:
        if debug.debug:
            self._log.debug(
                f"Waiting for the condition lock {threading.current_thread()}"
            )

        with self.cond:
            while self.state == external_state.RUNNING:
                self.cond.wait()

            if debug.debug:
                if self.state == external_state.EXITED:
                    self._log.debug(
                        f"Thread {self.thread} has exited from {threading.current_thread()}"
                    )
                elif self.state == external_state.PAUSED:
                    self._log.debug(
                        f"Thread {self.thread} has called yield from {threading.current_thread()}"
                    )

            if self.state == external_state.INIT:
                raise Exception(
                    f"Thread {self.thread} state was not allowed from {threading.current_thread()}"
                )

        return self.state



# Copyright cocotb contributors
# Licensed under the Revised BSD License, see LICENSE for details.
# SPDX-License-Identifier: BSD-3-Clause
import inspect
import os
import pdb
from typing import (
    Any,
    Callable,
    Coroutine,
    List,
    Optional,
    Union,
)

import cocotb
from cocotb._base_triggers import NullTrigger, Trigger
from cocotb._deprecation import deprecated
from cocotb._exceptions import InternalError
from cocotb._outcomes import Error, Outcome, Value
from cocotb._test_functions import TestSuccess
from cocotb.task import ResultType, Task

_pdb_on_exception = "COCOTB_PDB_ON_EXCEPTION" in os.environ


class RunningTest:
    """State of the currently executing Test."""

    # TODO
    # Make the tasks list a TaskManager.
    # Make shutdown errors and outcome be an ExceptionGroup from that TaskManager.
    # Replace result() with passing the outcome to the done callback.
    # Make this and Task the same object which is a Coroutine.
    # Reimplement the logic in the body of an async function.
    # Make RunningTest a normal Task that the RegressionManager runs and registers a
    #  done callback with.

    def __init__(
        self, test_complete_cb: Callable[[], None], main_task: Task[None]
    ) -> None:
        self._test_complete_cb: Callable[[], None] = test_complete_cb
        self._main_task: Task[None] = main_task
        self._main_task._add_done_callback(self._test_done_callback)

        self.tasks: List[Task[Any]] = [main_task]

        self._outcome: Union[None, Outcome[None]] = None
        self._shutdown_errors: list[Outcome[None]] = []

    def _test_done_callback(self, task: Task[None]) -> None:
        self.tasks.remove(task)
        # If cancelled, end the Test without additional error. This case would only
        # occur if a child threw a CancelledError or if the Test was forced to shutdown.
        if task.cancelled():
            self.abort(Value(None))
            return
        # Handle outcome appropriately and shut down the Test.
        e = task.exception()
        if e is None:
            self.abort(Value(task.result()))
        elif isinstance(e, TestSuccess):
            task._log.info("Test stopped early by this task")
            self.abort(Value(None))
        else:
            task._log.warning(e, exc_info=e)
            self.abort(Error(e))

    def start(self) -> None:
        cocotb._scheduler_inst._schedule_task_internal(self._main_task)
        cocotb._scheduler_inst._event_loop()

    def result(self) -> Outcome[None]:
        if self._outcome is None:  # pragma: no cover
            raise InternalError("Getting result before test is completed")

        if not isinstance(self._outcome, Error) and self._shutdown_errors:
            return self._shutdown_errors[0]
        return self._outcome

    def abort(self, outcome: Outcome[None]) -> None:
        """Force this test to end early."""

        # If we are shutting down, save any errors
        if self._outcome is not None:
            if isinstance(outcome, Error):
                self._shutdown_errors.append(outcome)
            return

        # Break into pdb on test end before all Tasks are killed.
        if _pdb_on_exception and isinstance(outcome, Error):
            try:
                pdb.post_mortem(outcome.error.__traceback__)
            except BaseException:
                pdb.set_trace()

        # Set outcome and cancel Tasks.
        self._outcome = outcome
        for task in self.tasks[:]:
            task._cancel_now()

        self._test_complete_cb()

    def add_task(self, task: Task[Any]) -> None:
        self.tasks.append(task)
        task._add_done_callback(self._task_done_callback)

    def _task_done_callback(self, task: Task[Any]) -> None:
        self.tasks.remove(task)
        # if cancelled, do nothing
        if task.cancelled():
            return
        # if there's a Task awaiting this one, don't fail
        if task.complete in cocotb._scheduler_inst._trigger2tasks:
            return
        # if no failure, do nothing
        e = task.exception()
        if e is None:
            return
        # there was a failure and no one is watching, fail test
        elif isinstance(e, TestSuccess):
            task._log.info("Test stopped early by this task")
            self.abort(Value(None))
        else:
            task._log.warning(e, exc_info=e)
            self.abort(Error(e))


def start_soon(
    coro: Union[Task[ResultType], Coroutine[Trigger, None, ResultType]],
    *,
    name: Optional[str] = None,
) -> Task[ResultType]:
    """
    Schedule a :term:`coroutine` to be run concurrently in a :class:`~cocotb.task.Task`.

    Note that this is not an :keyword:`async` function,
    and the new task will not execute until the calling task yields control.

    Args:
        coro: A :class:`!Task` or :term:`!coroutine` to be run concurrently.
        name:
            The task's name.

            .. versionadded:: 2.0

    Returns:
        The :class:`~cocotb.task.Task` that is scheduled to be run.

    .. versionadded:: 1.6
    """
    task = create_task(coro, name=name)
    cocotb._scheduler_inst._schedule_task(task)
    return task


@deprecated("Use ``cocotb.start_soon`` instead.")
async def start(
    coro: Union[Task[ResultType], Coroutine[Trigger, None, ResultType]],
    *,
    name: Optional[str] = None,
) -> Task[ResultType]:
    """
    Schedule a :term:`coroutine` to be run concurrently, then yield control to allow pending tasks to execute.

    The calling task will resume execution before control is returned to the simulator.

    When the calling task resumes, the newly scheduled task may have completed,
    raised an Exception, or be pending on a :class:`~cocotb.triggers.Trigger`.

    Args:
        coro: A :class:`!Task` or :term:`!coroutine` to be run concurrently.
        name:
            The task's name.

            .. versionadded:: 2.0

    Returns:
        The :class:`~cocotb.task.Task` that has been scheduled and allowed to execute.

    .. versionadded:: 1.6

    .. deprecated:: 2.0
        Use :func:`cocotb.start_soon` instead.
        If you need the scheduled Task to start before continuing the current Task,
        use an :class:`.Event` to block the current Task until the scheduled Task starts,
        like so:

        .. code-block:: python

            async def coro(started: Event) -> None:
                started.set()
                # Do stuff...


            task_started = Event()
            task = cocotb.start_soon(coro(task_started))
            await task_started.wait()
    """
    task = start_soon(coro, name=name)
    await NullTrigger()
    return task


def create_task(
    coro: Union[Task[ResultType], Coroutine[Trigger, None, ResultType]],
    *,
    name: Optional[str] = None,
) -> Task[ResultType]:
    """
    Construct a :term:`!coroutine` into a :class:`~cocotb.task.Task` without scheduling the task.

    The task can later be scheduled with :func:`cocotb.start` or :func:`cocotb.start_soon`.

    Args:
        coro: A :class:`!Task` or a :term:`!coroutine` to be turned into a :class:`!Task`.
        name:
            The task's name.

            .. versionadded:: 2.0

    Returns:
        Either the provided :class:`~cocotb.task.Task` or a new Task wrapping the coroutine.

    .. versionadded:: 1.6
    """
    if isinstance(coro, Task):
        if name is not None:
            coro.set_name(name)
        return coro
    elif isinstance(coro, Coroutine):
        task = Task[ResultType](coro, name=name)
        cocotb._regression_manager._running_test.add_task(task)
        return task
    elif inspect.iscoroutinefunction(coro):
        raise TypeError(
            f"Coroutine function {coro} should be called prior to being scheduled."
        )
    elif inspect.isasyncgen(coro):
        raise TypeError(
            f"{coro.__qualname__} is an async generator, not a coroutine. "
            "You likely used the yield keyword instead of await."
        )
    else:
        raise TypeError(
            f"Attempt to add an object of type {type(coro)} to the scheduler, "
            f"which isn't a coroutine: {coro!r}\n"
        )



# Copyright cocotb contributors
# Licensed under the Revised BSD License, see LICENSE for details.
# SPDX-License-Identifier: BSD-3-Clause

import os

debug: bool = bool(os.getenv("COCOTB_SCHEDULER_DEBUG"))
"""Global flag to enable additional debugging functionality.

Defaults to ``True`` if the :envvar:`COCOTB_SCHEDULER_DEBUG` environment variable is set,
but can be programmatically set by the user afterwards.

The ``"cocotb"`` logger should have its logging level set to :data:`logging.DEBUG`
to see additional debugging information in the test log.
This can be accomplished by setting the :envvar:`COCOTB_LOG_LEVEL` environment variable
to ``DEBUG``,
or using the following code.

.. code-block:: python

    import logging
    logging.getLogger("cocotb").setLevel(logging.DEBUG)

"""



# Copyright cocotb contributors
# Licensed under the Revised BSD License, see LICENSE for details.
# SPDX-License-Identifier: BSD-3-Clause

import functools
import inspect
import logging
import warnings
from itertools import product
from typing import (
    TYPE_CHECKING,
    Callable,
    Coroutine,
    Dict,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
    cast,
    overload,
)

from cocotb._base_triggers import Trigger
from cocotb._decorators import Test
from cocotb._typing import TimeUnit

if TYPE_CHECKING:
    from types import FrameType, FunctionType


class TestFactory:
    """Factory to automatically generate tests.

    Args:
        test_function: A Callable that returns the test Coroutine.
            Must take *dut* as the first argument.
        *args: Remaining arguments are passed directly to the test function.
            Note that these arguments are not varied. An argument that
            varies with each test must be a keyword argument to the
            test function.
        **kwargs: Remaining keyword arguments are passed directly to the test function.
            Note that these arguments are not varied. An argument that
            varies with each test must be a keyword argument to the
            test function.

    Assuming we have a common test function that will run a test. This test
    function will take keyword arguments (for example generators for each of
    the input interfaces) and generate tests that call the supplied function.

    This Factory allows us to generate sets of tests based on the different
    permutations of the possible arguments to the test function.

    For example, if we have a module that takes backpressure, has two configurable
    features where enabling ``feature_b`` requires ``feature_a`` to be active, and
    need to test against data generation routines ``gen_a`` and ``gen_b``:

    >>> tf = TestFactory(test_function=run_test)
    >>> tf.add_option(name="data_in", optionlist=[gen_a, gen_b])
    >>> tf.add_option("backpressure", [None, random_backpressure])
    >>> tf.add_option(
    ...     ("feature_a", "feature_b"), [(False, False), (True, False), (True, True)]
    ... )
    >>> tf.generate_tests()

    We would get the following tests:

        * ``gen_a`` with no backpressure and both features disabled
        * ``gen_a`` with no backpressure and only ``feature_a`` enabled
        * ``gen_a`` with no backpressure and both features enabled
        * ``gen_a`` with ``random_backpressure`` and both features disabled
        * ``gen_a`` with ``random_backpressure`` and only ``feature_a`` enabled
        * ``gen_a`` with ``random_backpressure`` and both features enabled
        * ``gen_b`` with no backpressure and both features disabled
        * ``gen_b`` with no backpressure and only ``feature_a`` enabled
        * ``gen_b`` with no backpressure and both features enabled
        * ``gen_b`` with ``random_backpressure`` and both features disabled
        * ``gen_b`` with ``random_backpressure`` and only ``feature_a`` enabled
        * ``gen_b`` with ``random_backpressure`` and both features enabled

    The tests are appended to the calling module for auto-discovery.

    Tests are simply named ``test_function_N``. The docstring for the test (hence
    the test description) includes the name and description of each generator.

    .. versionchanged:: 1.5
        Groups of options are now supported

    .. versionchanged:: 2.0
        You can now pass :func:`cocotb.test` decorator arguments when generating tests.

    .. deprecated:: 2.0
        Use :func:`cocotb.parametrize` instead.
    """

    def __init__(
        self,
        test_function: Callable[..., Coroutine[Trigger, None, None]],
        *args: object,
        **kwargs: object,
    ) -> None:
        warnings.warn(
            "TestFactory is deprecated, use `@cocotb.parametrize` instead",
            DeprecationWarning,
            stacklevel=2,
        )
        self.test_function = test_function
        self.args = args
        self.kwargs_constant = kwargs
        self.kwargs: Dict[
            Union[str, Sequence[str]],
            Union[Sequence[object], Sequence[Sequence[object]]],
        ] = {}
        self._log = logging.getLogger(f"TestFactory({self.test_function.__name__})")

    @overload
    def add_option(self, name: str, optionlist: Sequence[object]) -> None: ...

    @overload
    def add_option(
        self, name: Sequence[str], optionlist: Sequence[Sequence[object]]
    ) -> None: ...

    def add_option(
        self,
        name: Union[str, Sequence[str]],
        optionlist: Union[Sequence[object], Sequence[Sequence[object]]],
    ) -> None:
        """Add a named option to the test.

        Args:
            name:
                An option name, or an iterable of several option names. Passed to test as keyword arguments.

            optionlist:
                A list of possible options for this test knob.
                If N names were specified, this must be a list of N-tuples or
                lists, where each element specifies a value for its respective
                option.

        .. versionchanged:: 1.5
            Groups of options are now supported
        """
        if not isinstance(name, str):
            optionlist = cast("Sequence[Sequence[object]]", optionlist)
            for opt in optionlist:
                if len(name) != len(opt):
                    raise ValueError(
                        "Mismatch between number of options and number of option values in group"
                    )
        self.kwargs[name] = optionlist

    def generate_tests(
        self,
        *,
        prefix: Optional[str] = None,
        postfix: Optional[str] = None,
        stacklevel: int = 0,
        name: Optional[str] = None,
        timeout_time: Optional[float] = None,
        timeout_unit: TimeUnit = "step",
        expect_fail: bool = False,
        expect_error: Union[Type[BaseException], Tuple[Type[BaseException], ...]] = (),
        skip: bool = False,
        stage: int = 0,
    ) -> None:
        """
        Generate an exhaustive set of tests using the cartesian product of the
        possible keyword arguments.

        The generated tests are appended to the namespace of the calling
        module.

        Args:
            prefix:
                Text string to append to start of ``test_function`` name when naming generated test cases.
                This allows reuse of a single ``test_function`` with multiple :class:`TestFactories <.TestFactory>` without name clashes.

                .. deprecated:: 2.0
                    Use the more flexible ``name`` field instead.

            postfix:
                Text string to append to end of ``test_function`` name when naming generated test cases.
                This allows reuse of a single ``test_function`` with multiple :class:`TestFactories <.TestFactory>` without name clashes.

                .. deprecated:: 2.0
                    Use the more flexible ``name`` field instead.
            stacklevel:
                Which stack level to add the generated tests to. This can be used to make a custom TestFactory wrapper.

            name:
                Passed as ``name`` argument to :func:`cocotb.test`.

                .. versionadded:: 2.0

            timeout_time:
                Passed as ``timeout_time`` argument to :func:`cocotb.test`.

                .. versionadded:: 2.0

            timeout_unit:
                Passed as ``timeout_unit`` argument to :func:`cocotb.test`.

                .. versionadded:: 2.0

            expect_fail:
                Passed as ``expect_fail`` argument to :func:`cocotb.test`.

                .. versionadded:: 2.0

            expect_error:
                Passed as ``expect_error`` argument to :func:`cocotb.test`.

                .. versionadded:: 2.0

            skip:
                Passed as ``skip`` argument to :func:`cocotb.test`.

                .. versionadded:: 2.0

            stage:
                Passed as ``stage`` argument to :func:`cocotb.test`.

                .. versionadded:: 2.0
        """

        if prefix is not None:
            warnings.warn(
                "``prefix`` argument is deprecated. Use the more flexible ``name`` field instead.",
                DeprecationWarning,
                stacklevel=2,
            )
        else:
            prefix = ""

        if postfix is not None:
            warnings.warn(
                "``postfix`` argument is deprecated. Use the more flexible ``name`` field instead.",
                DeprecationWarning,
                stacklevel=2,
            )
        else:
            postfix = ""

        # trust the user puts a reasonable stacklevel in
        glbs = cast("FrameType", inspect.stack()[stacklevel][0].f_back).f_globals

        test_func_name = self.test_function.__qualname__ if name is None else name

        for index, testoptions in enumerate(
            dict(zip(self.kwargs, v)) for v in product(*self.kwargs.values())
        ):
            name = f"{prefix}{test_func_name}{postfix}_{(index + 1):03d}"
            doc: str = "Automatically generated test\n\n"

            # preprocess testoptions to split tuples
            testoptions_split: Dict[str, Sequence[object]] = {}
            for optname, optvalue in testoptions.items():
                if isinstance(optname, str):
                    optvalue = cast("Sequence[object]", optvalue)
                    testoptions_split[optname] = optvalue
                else:
                    # previously checked in add_option; ensure nothing has changed
                    optvalue = cast("Sequence[Sequence[object]]", optvalue)
                    assert len(optname) == len(optvalue)
                    for n, v in zip(optname, optvalue):
                        testoptions_split[n] = v

            for optname, optvalue in testoptions_split.items():
                if callable(optvalue):
                    optvalue = cast("FunctionType", optvalue)
                    if optvalue.__doc__ is None:
                        desc = "No docstring supplied"
                    else:
                        desc = optvalue.__doc__.split("\n")[0]
                    doc += f"\t{optname}: {optvalue.__qualname__} ({desc})\n"
                else:
                    doc += f"\t{optname}: {optvalue!r}\n"

            kwargs = self.kwargs_constant.copy()
            kwargs.update(testoptions_split)

            @functools.wraps(self.test_function)
            async def _my_test(dut: object, kwargs: Dict[str, object] = kwargs) -> None:
                await self.test_function(dut, *self.args, **kwargs)

            _my_test.__doc__ = doc
            _my_test.__name__ = name
            _my_test.__qualname__ = name

            if name in glbs:
                self._log.error(
                    "Overwriting %s in module %s. "
                    "This causes a previously defined testcase not to be run. "
                    "Consider using the `name`, `prefix`, or `postfix` arguments to augment the name.",
                    name,
                    glbs["__name__"],
                )

            test = Test(
                func=_my_test,
                name=name,
                module=glbs["__name__"],
                timeout_time=timeout_time,
                timeout_unit=timeout_unit,
                expect_fail=expect_fail,
                expect_error=expect_error,
                skip=skip,
                stage=stage,
            )

            glbs[test.name] = test



# Copyright cocotb contributors
# Copyright (c) 2013 Potential Ventures Ltd
# Copyright (c) 2013 SolarFlare Communications Inc
# Licensed under the Revised BSD License, see LICENSE for details.
# SPDX-License-Identifier: BSD-3-Clause

"""Utilities for implementors."""

import traceback
import types
from enum import Enum, IntEnum
from functools import update_wrapper, wraps
from types import TracebackType
from typing import (
    TYPE_CHECKING,
    Any,
    Iterable,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
    overload,
)

ExceptionTuple = Tuple[
    Type[BaseException], BaseException, TracebackType
]  # TypeAlias in Python 3.10


@overload
def remove_traceback_frames(
    tb_or_exc: ExceptionTuple, frame_names: List[str]
) -> ExceptionTuple: ...


@overload
def remove_traceback_frames(
    tb_or_exc: BaseException, frame_names: List[str]
) -> BaseException: ...


@overload
def remove_traceback_frames(
    tb_or_exc: TracebackType, frame_names: List[str]
) -> TracebackType: ...


def remove_traceback_frames(
    tb_or_exc: Union[ExceptionTuple, BaseException, TracebackType],
    frame_names: List[str],
) -> Union[ExceptionTuple, BaseException, TracebackType]:
    """
    Strip leading frames from a traceback

    Args:
        tb_or_exc:
            Object to strip frames from. If an exception is passed, creates
            a copy of the exception with a new shorter traceback. If a tuple
            from `sys.exc_info` is passed, returns the same tuple with the
            traceback shortened
        frame_names:
            Names of the frames to strip, which must be present at the top of the Traceback or Exception.

    Returns:
        Traceback or Exception passed to the function with the *frame_names* stripped out.
    """
    # self-invoking overloads
    if isinstance(tb_or_exc, BaseException):
        exc: BaseException = tb_or_exc
        return exc.with_traceback(
            remove_traceback_frames(
                cast("TracebackType", exc.__traceback__), frame_names
            )
        )
    elif isinstance(tb_or_exc, tuple):
        exc_type, exc_value, exc_tb = tb_or_exc
        exc_tb = remove_traceback_frames(exc_tb, frame_names)
        return exc_type, exc_value, exc_tb
    # base case
    else:
        tb: TracebackType = tb_or_exc
        for frame_name in frame_names:
            # the assert and cast are there assuming the frame_names being removed are correct
            assert tb.tb_frame.f_code.co_name == frame_name
            tb = cast("TracebackType", tb.tb_next)
        return tb


def walk_coro_stack(
    coro: "types.CoroutineType[Any, Any, Any]",
) -> Iterable[Tuple[types.FrameType, int]]:
    """Walk down the coroutine stack, starting at *coro*.

    Args:
        coro: The :class:`coroutine` object to traverse.

    Yields:
        Frame and line number of each frame in the coroutine.
    """
    c: Optional[types.CoroutineType[Any, Any, Any]] = coro
    while c is not None:
        try:
            f = c.cr_frame
        except AttributeError:
            break
        else:
            c = c.cr_await
        if f is not None:
            yield (f, f.f_lineno)


def extract_coro_stack(
    coro: "types.CoroutineType[Any, Any, Any]", limit: Optional[int] = None
) -> traceback.StackSummary:
    r"""Create a list of pre-processed entries from the coroutine stack.

    This is based on :func:`traceback.extract_tb`.

    If *limit* is omitted or ``None``, all entries are extracted.
    The list is a :class:`traceback.StackSummary` object, and
    each entry in the list is a :class:`traceback.FrameSummary` object
    containing attributes ``filename``, ``lineno``, ``name``, and ``line``
    representing the information that is usually printed for a stack
    trace. The line is a string with leading and trailing
    whitespace stripped; if the source is not available it is ``None``.

    Args:
        coro: The :class:`coroutine` object from which to extract a stack.
        level: The maximum number of frames from *coro*\ s stack to extract.

    Returns:
        The stack of *coro*.
    """
    return traceback.StackSummary.extract(walk_coro_stack(coro), limit=limit)


EnumT = TypeVar("EnumT", bound=Enum)


class DocEnum(Enum):
    """Like :class:`enum.Enum`, but allows documenting enum values.

    Documentation for enum members can be optionally added by setting enum values to a tuple of the intended value and the docstring.
    This adds the provided docstring to the ``__doc__`` field of the enum value.

    .. code-block:: python

        class MyEnum(DocEnum):
            \"\"\"Class documentation\"\"\"

            VALUE1 = 1, "Value documentation"
            VALUE2 = 2  # no documentation

    Taken from :ref:`this StackOverflow answer <https://stackoverflow.com/questions/50473951/how-can-i-attach-documentation-to-members-of-a-python-enum/50473952#50473952>`
    by :ref:`Eric Wieser <https://stackoverflow.com/users/102441/eric>`,
    as recommended by the ``enum_tools`` documentation.
    """

    def __new__(cls: Type[EnumT], value: object, doc: Optional[str] = None) -> EnumT:
        # super().__new__() assumes the value is already an enum value
        # so we side step that and create a raw object and fill in _value_
        self = object.__new__(cls)
        self._value_ = value
        if doc is not None:
            self.__doc__ = doc
        return self


IntEnumT = TypeVar("IntEnumT", bound=IntEnum)


class DocIntEnum(IntEnum):
    """Like DocEnum but for :class:`IntEnum` enum types."""

    def __new__(cls: Type[IntEnumT], value: int, doc: Optional[str] = None) -> IntEnumT:
        self = int.__new__(cls, value)
        self._value_ = value
        if doc is not None:
            self.__doc__ = doc
        return self


if TYPE_CHECKING:
    F = TypeVar("F")

    def cached_method(f: F) -> F: ...

else:

    class cached_method:
        def __init__(self, method):
            self._method = method
            update_wrapper(self, method)

        def __get__(self, instance, objtype=None):
            if instance is None:
                return self

            cache = {}

            @wraps(self._method)
            def lookup(*args, **kwargs):
                key = (args, tuple(kwargs.items()))
                try:
                    return cache[key]
                except KeyError:
                    res = self._method(instance, *args, **kwargs)
                    cache[key] = res
                    return res

            lookup.cache = cache

            setattr(instance, self._method.__name__, lookup)
            return lookup

        def __call__(self, instance, *args, **kwargs):
            func = getattr(instance, self._method.__name__)
            return func(*args, **kwargs)


T = TypeVar("T")


if TYPE_CHECKING:

    def singleton(orig_cls: T) -> T: ...

else:

    def singleton(orig_cls):
        """Class decorator which turns a type into a Singleton type."""
        orig_new = orig_cls.__new__
        orig_init = orig_cls.__init__
        instance = None

        @wraps(orig_cls.__new__)
        def __new__(cls, *args, **kwargs):
            nonlocal instance
            if instance is None:
                instance = orig_new(cls, *args, **kwargs)
                orig_init(instance, *args, **kwargs)
            return instance

        @wraps(orig_cls.__init__)
        def __init__(self, *args, **kwargs):
            pass

        orig_cls.__new__ = __new__
        orig_cls.__init__ = __init__
        return orig_cls


def pointer_str(obj: object) -> str:
    """Get the memory address of *obj* as used in :meth:`object.__repr__`.

    This is equivalent to ``sprintf("%p", id(obj))``, but Python does not
    support ``%p``.
    """
    full_repr = object.__repr__(obj)  # gives "<{type} object at {address}>"
    return full_repr.rsplit(" ", 1)[1][:-1]


def safe_divide(a: float, b: float) -> float:
    """Used when computing time ratios to ensure no exception is raised if either time is 0."""
    try:
        return a / b
    except ZeroDivisionError:
        if a == 0:
            return float("nan")
        else:
            return float("inf")



# Copyright cocotb contributors
# Copyright (c) 2013 Potential Ventures Ltd
# Copyright (c) 2013 SolarFlare Communications Inc
# Licensed under the Revised BSD License, see LICENSE for details.
# SPDX-License-Identifier: BSD-3-Clause

"""A collection of triggers which a testbench can :keyword:`await`."""

import warnings
from decimal import Decimal
from fractions import Fraction
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Generic,
    Optional,
    TypeVar,
    Union,
)

import cocotb
import cocotb.handle
from cocotb import simulator
from cocotb._base_triggers import Trigger
from cocotb._deprecation import deprecated
from cocotb._typing import RoundMode, TimeUnit
from cocotb._utils import pointer_str, singleton
from cocotb.utils import get_sim_steps, get_time_from_sim_steps

if TYPE_CHECKING:
    from cocotb._py_compat import Self


class GPITrigger(Trigger):
    """A trigger for a simulation event."""

    def __init__(self) -> None:
        super().__init__()
        self._cbhdl: Optional[simulator.gpi_cb_hdl] = None

    def _unprime(self) -> None:
        """Disable a primed trigger, can be re-primed."""
        if self._cbhdl is not None:
            self._cbhdl.deregister()
        return super()._unprime()

    def _cleanup(self) -> None:
        self._cbhdl = None
        return super()._cleanup()


class Timer(GPITrigger):
    r"""Fire after the specified simulation time period has elapsed.

    This trigger will *always* consume some simulation time
    and will return control to the :keyword:`await`\ ing task at the beginning of the time step.

    Args:
        time: The time value.

            .. versionchanged:: 1.5
                Previously this argument was misleadingly called `time_ps`.

        unit: The unit of the time value.

            One of ``'step'``, ``'fs'``, ``'ps'``, ``'ns'``, ``'us'``, ``'ms'``, ``'sec'``.
            When *unit* is ``'step'``,
            the timestep is determined by the simulator (see :make:var:`COCOTB_HDL_TIMEPRECISION`).

            .. versionchanged:: 2.0
                Renamed from ``units``.

        round_mode:

            String specifying how to handle time values that sit between time steps
            (one of ``'error'``, ``'round'``, ``'ceil'``, ``'floor'``, ``None``).
            A ``None`` argument is converted to the current value of :attr:`.Timer.round_mode`.

    Raises:
        ValueError: If a non-positive value is passed for Timer setup.

    Usage:
        >>> await Timer(100, unit="ps")

        The time can also be a ``float``:

        >>> await Timer(100e-9, unit="sec")

        which is particularly convenient when working with frequencies:

        >>> freq = 10e6  # 10 MHz
        >>> await Timer(1 / freq, unit="sec")

        Other built-in exact numeric types can be used too:

        >>> from fractions import Fraction
        >>> await Timer(Fraction(1, 10), unit="ns")

        >>> from decimal import Decimal
        >>> await Timer(Decimal("100e-9"), unit="sec")

        These are most useful when using computed durations while
        avoiding floating point inaccuracies.

    .. versionchanged:: 1.5
        Raise an exception when Timer uses a negative value as it is undefined behavior.
        Warn for 0 as this will cause erratic behavior in some simulators as well.

    .. versionchanged:: 1.5
        Support ``'step'`` as the *unit* argument to mean "simulator time step".

    .. versionchanged:: 1.6
        Support rounding modes.

    .. versionremoved:: 2.0
        Passing ``None`` as the *unit* argument was removed, use ``'step'`` instead.

    .. versionremoved:: 2.0
        The ``time_ps`` parameter was removed, use the ``time`` parameter instead.

    .. versionchanged:: 2.0
        Passing ``0`` as the *time* argument now raises a :exc:`ValueError`.
    """

    round_mode: ClassVar[RoundMode] = "error"
    """The default rounding mode."""

    def __init__(
        self,
        time: Union[float, Fraction, Decimal],
        unit: TimeUnit = "step",
        *,
        round_mode: Optional[RoundMode] = None,
        units: None = None,
    ) -> None:
        super().__init__()
        if time <= 0:
            raise ValueError("Timer argument time must be positive")
        if units is not None:
            warnings.warn(
                "The 'units' argument has been renamed to 'unit'.",
                DeprecationWarning,
                stacklevel=2,
            )
            unit = units
        if round_mode is None:
            round_mode = type(self).round_mode
        self._sim_steps = get_sim_steps(time, unit, round_mode=round_mode)
        # If we round to 0, we fix it up to 1 step as rounding is imprecise,
        # and Timer(0) is invalid.
        if self._sim_steps == 0:
            self._sim_steps = 1

    def _prime(self, callback: Callable[["Self"], None]) -> None:
        """Register for a timed callback."""
        if self._cbhdl is None:
            self._cbhdl = simulator.register_timed_callback(
                self._sim_steps, callback, self
            )
            if self._cbhdl is None:
                raise RuntimeError(f"Unable set up {self!s} Trigger")
        super()._prime(callback)

    def __repr__(self) -> str:
        return "<{} of {:1.2f}ps at {}>".format(
            type(self).__qualname__,
            get_time_from_sim_steps(self._sim_steps, unit="ps"),
            pointer_str(self),
        )


@singleton
class ReadOnly(GPITrigger):
    """Fires when the current simulation timestep moves to the read-only phase.

    The read-only phase is entered when the current timestep no longer has any further delta steps.
    This will be a point where all the signal values are stable as there are no more RTL events scheduled for the timestep.
    The simulator will not allow scheduling of more events in this timestep.
    Useful for monitors which need to wait for all processes to execute (both RTL and cocotb) to ensure sampled signal values are final.
    """

    def _prime(self, callback: Callable[["Self"], None]) -> None:
        if isinstance(current_gpi_trigger(), ReadOnly):
            raise RuntimeError(
                "Attempted illegal transition: awaiting ReadOnly in ReadOnly phase"
            )
        if self._cbhdl is None:
            self._cbhdl = simulator.register_readonly_callback(callback, self)
            if self._cbhdl is None:
                raise RuntimeError(f"Unable set up {self!s} Trigger")
        super()._prime(callback)

    def __repr__(self) -> str:
        return f"{type(self).__qualname__}()"


@singleton
class ReadWrite(GPITrigger):
    """Fires when the read-write simulation phase is reached."""

    def _prime(self, callback: Callable[["Self"], None]) -> None:
        if isinstance(current_gpi_trigger(), ReadOnly):
            raise RuntimeError(
                "Attempted illegal transition: awaiting ReadWrite in ReadOnly phase"
            )
        if self._cbhdl is None:
            self._cbhdl = simulator.register_rwsynch_callback(callback, self)
            if self._cbhdl is None:
                raise RuntimeError(f"Unable set up {self!s} Trigger")
        super()._prime(callback)

    def __repr__(self) -> str:
        return f"{type(self).__qualname__}()"


@singleton
class NextTimeStep(GPITrigger):
    """Fires when the next time step is started."""

    def _prime(self, callback: Callable[["Self"], None]) -> None:
        if self._cbhdl is None:
            self._cbhdl = simulator.register_nextstep_callback(callback, self)
            if self._cbhdl is None:
                raise RuntimeError(f"Unable set up {self!s} Trigger")
        super()._prime(callback)

    def __repr__(self) -> str:
        return f"{type(self).__qualname__}()"


_SignalType = TypeVar("_SignalType", bound="cocotb.handle.ValueObjectBase[Any, Any]")


class _EdgeBase(GPITrigger, Generic[_SignalType]):
    """Internal base class that fires on a given edge of a signal."""

    _edge_type: ClassVar[int]
    signal: _SignalType

    @classmethod
    def _make(cls, signal: _SignalType) -> "Self":
        self = GPITrigger.__new__(cls)
        GPITrigger.__init__(self)
        self.signal = signal
        return self

    def __init__(self, _: _SignalType) -> None:
        pass

    def _prime(self, callback: Callable[["Self"], None]) -> None:
        if self._cbhdl is None:
            self._cbhdl = simulator.register_value_change_callback(
                self.signal._handle, callback, type(self)._edge_type, self
            )
            if self._cbhdl is None:
                raise RuntimeError(f"Unable set up {self!s} Trigger")
        super()._prime(callback)

    def __repr__(self) -> str:
        return f"{type(self).__qualname__}({self.signal!r})"


class RisingEdge(_EdgeBase["cocotb.handle.LogicObject"]):
    """Fires on the rising edge of *signal*, on a transition to ``1``.

    Only valid for scalar ``logic`` or ``bit``-typed signals.

    Args:
        signal: The signal upon which to wait for a rising edge.

    Raises:
        TypeError: If *signal* is not a 1-bit ``logic`` or ``bit``-typed object.

    .. note::
        Prefer :attr:`await signal.rising_edge <cocotb.handle.LogicObject.rising_edge>` to ``await RisingEdge(signal)``.

    .. warning::
        On many simulators this will trigger on transitions from non-``0``/``1`` value to ``1``,
        not just from ``0`` to ``1`` like the ``rising_edge`` function in VHDL.
    """

    _edge_type = simulator.RISING

    def __new__(cls, signal: "cocotb.handle.LogicObject") -> "RisingEdge":
        if not (isinstance(signal, cocotb.handle.LogicObject)):
            raise TypeError(
                f"{cls.__qualname__} requires a scalar LogicObject. Got {signal!r} of type {type(signal).__qualname__}"
            )
        return signal.rising_edge


class FallingEdge(_EdgeBase["cocotb.handle.LogicObject"]):
    """Fires on the falling edge of *signal*, on a transition to ``0``.

    Only valid for scalar ``logic`` or ``bit``-typed signals.

    Args:
        signal: The signal upon which to wait for a rising edge.

    Raises:
        TypeError: If *signal* is not a 1-bit ``logic`` or ``bit``-typed object.

    .. note::
        Prefer :attr:`await signal.falling_edge <cocotb.handle.LogicObject.falling_edge>` to ``await FallingEdge(signal)``.

    .. warning::
        On many simulators this will trigger on transitions from non-``0``/``1`` value to ``0``,
        not just from ``1`` to ``0`` like the ``falling_edge`` function in VHDL.
    """

    _edge_type = simulator.FALLING

    def __new__(cls, signal: "cocotb.handle.LogicObject") -> "FallingEdge":
        if not (isinstance(signal, cocotb.handle.LogicObject)):
            raise TypeError(
                f"{cls.__qualname__} requires a scalar LogicObject. Got {signal!r} of type {type(signal).__qualname__}"
            )
        return signal.falling_edge


class ValueChange(_EdgeBase["cocotb.handle._NonIndexableValueObjectBase[Any, Any]"]):
    """Fires on any value change of *signal*.

    Args:
        signal: The signal upon which to wait for a value change.

    Raises:
        TypeError: If the signal is not an object which can change value.

    .. note::
        Prefer :attr:`await signal.value_change <cocotb.handle.NonArrayValueObject.value_change>` to ``await ValueChange(signal)``.

    .. versionadded:: 2.0
    """

    _edge_type = simulator.VALUE_CHANGE

    def __new__(
        cls, signal: "cocotb.handle._NonIndexableValueObjectBase[Any, Any]"
    ) -> "ValueChange":
        if not isinstance(signal, cocotb.handle._NonIndexableValueObjectBase):
            raise TypeError(
                f"{cls.__qualname__} requires a simulation object derived from ValueObjectBase. "
                f"Got {signal!r} of type {type(signal).__qualname__}"
            )
        return signal.value_change


class Edge(ValueChange):
    """Fires on any value change of *signal*.

    Args:
        signal: The signal upon which to wait for a value change.

    Raises:
        TypeError: If the signal is not an object which can change value.

    .. deprecated:: 2.0

        Use :attr:`signal.value_change <cocotb.handle.NonArrayValueObject.value_change>` instead.
    """

    @deprecated("Use `signal.value_change` instead.")
    def __new__(
        cls, signal: "cocotb.handle._NonIndexableValueObjectBase[Any, Any]"
    ) -> "Edge":
        if not isinstance(signal, cocotb.handle._NonIndexableValueObjectBase):
            raise TypeError(
                f"{cls.__qualname__} requires a simulation object derived from ValueObjectBase. "
                f"Got {signal!r} of type {type(signal).__qualname__}"
            )
        return signal._edge


# The initializer is a lie, but a useful one. Perhaps one day this can be something like `StartupTrigger`.`
_current_gpi_trigger = Timer(1, "step")  # type: Union[None, GPITrigger]


def current_gpi_trigger() -> GPITrigger:
    """Return the last GPITrigger that fired."""
    if _current_gpi_trigger is None:
        raise RuntimeError("No GPI trigger has fired.")
    return _current_gpi_trigger



# Copyright cocotb contributors
# Licensed under the Revised BSD License, see LICENSE for details.
# SPDX-License-Identifier: BSD-3-Clause
"""Checks if a JUnit results file exists and whether there was failing tests."""

import argparse
import sys
from pathlib import Path
from typing import Tuple
from xml.etree import ElementTree


def get_results(results_xml_file: Path) -> Tuple[int, int]:
    """Return number of tests and fails in *results_xml_file*.

    Returns:
        Tuple of number of tests and number of fails.

    Raises:
        RuntimeError: If *results_xml_file* is non-existent.
    """

    __tracebackhide__ = True  # Hide the traceback when using PyTest.

    if not results_xml_file.is_file():
        raise RuntimeError(
            f"ERROR: Simulation terminated abnormally. Results file {results_xml_file} not found."
        )

    num_tests = 0
    num_failed = 0

    tree = ElementTree.parse(results_xml_file)
    for ts in tree.iter("testsuite"):
        for tc in ts.iter("testcase"):
            num_tests += 1
            for _ in tc.iter("failure"):
                num_failed += 1

    return (num_tests, num_failed)


def _get_parser() -> argparse.ArgumentParser:
    """Return the cmdline parser"""
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "results_file", help="Path to XML file holding JUnit test results.", type=Path
    )
    return parser


def main() -> int:
    parser = _get_parser()
    args = parser.parse_args()

    try:
        (_, num_failed) = get_results(args.results_file)
    except RuntimeError:
        return 1
    return num_failed


if __name__ == "__main__":
    rc = main()
    sys.exit(rc)
